#!/usr/bin/env python

"""
This is a wrangler for RADICAL-Pilot (RP) sessions. It assumes the following:

* One RP session corresponds to one experiment run.
* One experiment has multiple runs.
* Profiles and json file have been collected for each RP session.
* Data for each session are collected into a directory with the same name as
  the session ID. For example: 'rp.session.titan-ext1.merzky1.017242.0012'.
* All the session directories are collected into a single experiment directory.
* All the experiment directories are collected within a single data directory.

The wrangler cycles through all the experiment directories within the data
directory and calculates all the durations for each session, pilot, and unit.
Durations and associated timestamps are saved into three csv files:
sessions.csv for the session entities; pilots.csv for pilot entities; and
units.csv for unit entities.

CSV files are written incrementally

Examples:

* Typically, data are organized in a directory trhee similar to:
  data/exp1/rp.session.titan-ext1.merzky1.017242.0012

Todo:

1. Intercept and handle more errors from ra:
  - ValueError: no duration defined for given constraints
2. Eliminate records in units/pilots when session fails to wrangle
   (transactions).

"""

import os
import sys
import glob
import fcntl
import getopt
import psutil
import numpy as np
import pandas as pd
import multiprocessing as mp

import radical.utils     as ru
import radical.pilot     as rp
import radical.analytics as ra

from sqlalchemy import create_engine


# ------------------------------------------------------------------------------
#
def help():
        message = """
        -d --ddir       Name of the directory containing all the experiment
                        directories. For example '../data'.
        -t --etag       Name of the tag identifying each experiment directory.
                        For example, given three experiment directories exp1,
                        exp2, exp3, the experiment tag is 'exp'.
        -e --eid        Number of the experiment to wrangle. For example '1'.
                        When specified, the wrangler wrangles only the
                        sessions wihtin this experiment directory -t+-e (exp1).
        -s --sid        Name of the directory of a session of RADICAL Pilot.
                        For example 'rp.session.radical.mturilli.017233.0002'.
                        When specified, the wrangler wrangles only this
                        session.
        -o --odir       Name of the directory where to load and save the csv
                        files created by the wrangler. When not specified,
                        -d is used.
        -h --help       Prints the help page.
        -u --usage      Prints usage command.
        """
        return usage() + message


# ------------------------------------------------------------------------------
#
def usage():
        message = """
        ra-wrangler.py -d <directory> -t <tag>
                       [-e <integer>] [-s <rp_session_ID>][-o <directory>]
                       [-h] [-u]
        """
        return message


# ------------------------------------------------------------------------------
#
def clparse(argv):
    clopts = {'ddir': None,  # data directory (mandatory).
              'eid' : None,  # experiment tag (mandatory).
              'enum': None,  # experiment number.
              'sid' : None,  # session ID.
              'odir': None}  # directory where to save csv files.

    try:
        opts, args = getopt.getopt(argv, 'hud:t:e:s:o:',
            ['help','usage','ddir=','etag=','eid=','sid=','odir='])
        if not opts:
            print('No options supplied')
            print(usage())
            sys.exit(1)

    except getopt.GetoptError as e:
        print(e)
        usage()
        sys.exit(2)

    for opt, arg in opts:

        if opt in ['-h', '--help']:
            print(help())
            sys.exit(0)

        if   opt in ('-d', '--edir'): clopts['ddir'] = arg
        elif opt in ('-t', '--etag'): clopts['etag'] = arg
        elif opt in ('-e', '--eid') : clopts['eid']  = arg
        elif opt in ('-s', '--sid') : clopts['sid']  = arg
        elif opt in ('-o', '--odir'): clopts['odir'] = arg
        else: raise ValueError('invalid option %s' % opt)


    # Define the directory where to output the cvs files created by the
    # wrangler.
    if not clopts['odir']:
        clopts['odir'] = clopts['ddir']

    # Check for mandatory arguments
    if not clopts['ddir'] or not clopts['etag']:
        print('One or more mandatory option was not supplied')
        print(usage())
        sys.exit(3)

    return clopts


# ------------------------------------------------------------------------------
#
def initialize_entity(etype=None):

    # Columns for each entity's csv
    entities = {
        'session': {'sid'          : [],     # Session ID
                    'session'      : [],     # RA session objects
                    'experiment'   : [],     # Experiment ID
                    'nhost'        : [],     # #host for CU execution
                    'nunit'        : [],     # #units
                    'nunit_done'   : [],     # #active units
                    'nunit_failed' : [],     # #failed units
                    'npilot'       : [],     # #pilots
                    'npilot_active': [],     # #active pilots
                    'ncore'        : [],     # #cores
                    'ncore_active' : []},    # #active cores
        'pilot'  : {'pid'          : [],     # Pilot ID
                    'sid'          : [],     # Session ID
                    'hid'          : [],     # Host ID
                    'ncore'        : [],     # #cores
                    'nunit'        : [],     # #units executed
                    'experiment'   : []},    # Experiment ID
        'unit'   : {'uid'          : [],     # Unit ID
                    'sid'          : [],     # Session ID
                    'pid'          : [],     # Pilot ID
                    'hid'          : [],     # Host ID
                    'experiment'   : []}}    # Experiment ID

    # Add the duration label of each state of each entity.
    for duration in pdm:
        entities['session'][duration] = list()
        entities['pilot'][duration]   = list()

    for duration in udm:
        entities['session'][duration] = list()
        entities['unit'][duration]    = list()

    if etype not in ['session', 'pilot', 'unit']:
        print('Cannot itialize entity %s' % etype)
        sys.exit(1)

    # Return the empty data structure of the requested entity.
    return entities[etype]


# ------------------------------------------------------------------------------
#
def load_df(etype=None, sid=None):

    if etype not in ['session', 'pilot', 'unit']:
        print('Cannot itialize entity %s' % etype)
        sys.exit(1)

    # Initialize an empty DF with the entity's properties.
    df = pd.DataFrame(initialize_entity(etype=etype))

    # Load the entity's csv into a Panda DataFrame.
    if os.path.isfile(csvs[etype]):
        df = pd.read_csv(csvs[etype], index_col=0)

        # Prune the DF to save memory.
        if sid:
            df = df[df.sid == sid]

    return df


# ------------------------------------------------------------------------------
#
def store_df(new_df, stored=pd.DataFrame(), etype=None):

    # skip storing if no new data are passed.
    if new_df.empty:
        print('WARNING: attempting to store an empty DF (skip).')
        return

    # Set header for csv file
    header = True
    if os.path.isfile(csvs[etype]):
        header = False

    if etype == 'session':
        new_sessions = new_df.drop('session', axis=1)
        if stored.empty:
            sessions = new_sessions
        else:
            sessions = stored.append(new_sessions)
        sessions.to_csv(csvs[etype])

    elif etype not in ['pilot', 'unit']:
        print('Cannot store DF to %s' % etype)
        sys.exit(1)

    if stored.empty:
        df = new_df

    else:
        df = stored.append(new_df, ignore_index=True)

    df.reset_index(inplace=True, drop=True)

    with open(csvs[etype], 'a') as f:
        fcntl.flock(f, fcntl.LOCK_EX)
        df.to_csv(f, header=header)
        fcntl.flock(f, fcntl.LOCK_UN)


# ------------------------------------------------------------------------------
#
def parse_osg_hostid(hostid):
    '''
    Heuristic: eliminate node-specific information from hostID.
    '''
    domain = None

    # Split domain name from IP.
    host = hostid.split(':')

    # Split domain name into words.
    words = host[0].split('.')

    # Get the words in the domain name that do not contain
    # numbers. Most hostnames have no number but there are
    # exceptions.
    literals = [l for l in words 
                  if not any((number in set('0123456789')) for number in l)]

    # Check for exceptions:
    # a. every word of the domain name has a number
    if len(literals) == 0:

        # Some hostname use '-' instead of '.' as word separator.
        # The parser would have returned a single word and the
        # any of that word may have a number.
        if '-' in host[0]:
            words    = host[0].split('-')
            literals = [l for l in words if not
                        any((number in set('0123456789')) for number in l)]

            # FIXME: We do not check the size of literals.
            domain = '.'.join(literals)

        # Some hostnames may have only the name of the node. We
        # have to keep the IP to decide later on whether two nodes
        # are likely to belong to the same cluster.
        elif 'nod' in host[0]:
            domain = '.'.join(host)

        # FIXME: ad hoc parsing
        elif 'n0' in host[0]:
            domain = 'n0x.10.2.x.x'

        # The hostname is identified by an alphanumeric string
        else:
            domain = '.'.join(host)

    # Some hostnames DO have numbers in their name.
    elif len(literals) == 1:
        domain = '.'.join(words[1:])

    # Some hostname are just simple to parse.
    else:
        domain = '.'.join(literals)

    # FIXME: When everything else fails, ad hoc manipulations of domain string.
    if   'its.osg' in domain: domain = 'its.osg'
    elif 'nodo'    in domain: domain = 'nodo'
    elif 'bu.edu'  in domain: domain = 'bu.edu'

    return domain


# ------------------------------------------------------------------------------
#
def load_pilots(sid, exp, sra_pilots, pdm, pu_rels, pts):

    sys.stdout.write('\n%s --- %s' % (exp, sid))
    ps = initialize_entity(etype='pilot')

    # Did we already store pilots of this session?
    stored_pilots = load_df(etype='pilot', sid=sid)
    stored_pids   = list()

    if stored_pilots['sid'].any():
        stored_pilots_sid = stored_pilots.loc[stored_pilots['sid'] == sid].copy()
        stored_pids       = stored_pilots_sid['pid'].values.tolist()

    # Derive properties and duration for each pilot.
    for pid in sorted(sra_pilots.list('uid')):

        # Skip session if its pilots have been already stored.
        if pid in stored_pids:
            sys.stdout.write('\n%s already in %s' % (pid, csvs['pilot']))
            continue

        # Pilot properties.
        sys.stdout.write('\n' + pid + ':\n')
        ps['pid'].append(pid)
        ps['sid'].append(sid)
        ps['experiment'].append(exp)

        # Get pilot entity from RA session.
        pentity = sra_pilots.get(uid=pid)[0]

        # Host ID.
        if pentity.cfg['hostid']:
            ps['hid'].append(parse_osg_hostid(pentity.cfg['hostid']))
        else:
            ps['hid'].append(np.nan)

        # Number of cores of the pilot.
        ps['ncore'].append(pentity.description['cores'])

        # Number of units executed.
        ps['nunit'].append(len(pu_rels[pid]))

        # Pilot Timestamps.
        for state in pts:
            if state not in ps:
                ps[state] = list()

            try:
                ps[state].append(pentity.timestamps(state=state)[0])
            except:
                if state not in rp.FINAL:
                    print('WARNING: no timestamp for state %s' % state)
                ps[state].append(np.nan)

        # Pilot durations.
        for duration in pdm:

            # TODO: AM: why are not all inserted here?
            #           Can there be more than one (why is it a list)?
            if duration not in ps:
                ps[duration] = list()

            try:
                ps[duration].append(pentity.duration(event=pdm[duration]))
                sys.stdout.write(' %s' % duration)
            except:
                print(' WARNING: Failed to calculate duration %s' % duration)
                ps[duration].append(np.nan)

    # Store pilots DF to csv and reload into memory to return the complete
    # DF for the given sid.
    if ps['pid']:
        pilots = pd.DataFrame(ps)
        # FIXME: This is a workaround to load the full DF before saving it. The
        # DF is then unloaded loading only the portion related to the pilots
        # just wrangled. This avoids using memory but needs cleanup.
        store_df(pilots, etype='pilot')
        stored_pilots = load_df(etype='pilot', sid=sid)
        print('\nstored in %s.' % csvs['pilot'])

    # Returns the DF of the stored pilots if no new pilots have been added;
    # the DF with the old and new pilots otherwise.
    return stored_pilots


# ------------------------------------------------------------------------------
#
def load_units(sid, exp, sra_units, udm, pilots, sra, pu_rels, uts):

    sys.stdout.write('\n%s --- %s' % (exp, sid))
    us = initialize_entity(etype='unit')

    # Did we already store units of this session?
    stored_units = load_df(etype='unit', sid=sid)
    stored_uids  = list()

    if stored_units['sid'].any():
        stored_units_sid = stored_units.loc[stored_units['sid'] == sid].copy()
        stored_uids = stored_units_sid['uid'].values.tolist()

    # Derive properties and duration for each unit.
    for uid in sorted(sra_units.list('uid')):

        # Skip session if its pilots have been already stored.
        if uid in stored_uids:
            sys.stdout.write('\n%s already stored in %s' % (uid, csvs['unit']))
            continue

        # Properties.
        sys.stdout.write('\n' + uid + ':\n')
        us['uid'].append(uid)
        us['sid'].append(sid)
        us['experiment'].append(exp)

        # Get unit entity from RA session
        uentity = sra_units.get(uid=uid)[0]

        # Unit Timestamps.
        for state in uts:
            if state not in us:
                us[state] = list()

            try:
                us[state].append(uentity.timestamps(state=state)[0])
            except:
                if state not in ['CANCELED', 'FAILED']:
                    print('WARNING: Failed to get timestampe for %s' % state)
                us[state].append(np.nan)

        # Durations.
        for duration in udm:
            if duration not in us:
                us[duration] = list()
            try:
                # TODO: this is a temporary fix for inconsistent state model.
                # TODO: AM: WHAT INCONSISTENT STATE MODEL?? ;)  Still needed?
                if duration == 'U_AGENT_EXECUTING':
                    if 'AGENT_STAGING_OUTPUT_PENDING' in uentity.states and \
                       'FAILED'                       in uentity.states:
                        us[duration].append(np.nan)
                        continue

                us[duration].append(uentity.duration(event=udm[duration]))
                sys.stdout.write(' %s' % duration)

            except:
                print('\nWARNING: Failed to calculate duration %s' % \
                    duration)
                us[duration].append(np.nan)

        # pilot and host on which the unit has been executed.
        punit = [key[0] for key in list(pu_rels.items())
                        if  uid in key[1]]
        if punit:
            punit = punit[0]
        else:
            print('\nDEBUG: pu_rels.items():\n' % list(pu_rels.items()))
            print('WARNING: empty pilot name for unit %s' % uid)
            punit = None

        hid = pilots[(pilots['sid'] == sid) &
                     (pilots['pid'] == punit)]['hid'].tolist()[0]
        
        us['pid'].append(punit)
        us['hid'].append(hid)

    # Store unit DF to csv and reload into memory to return the complete
    # DF for the given sid.
    if us['pid']:
        units = pd.DataFrame(us)
        # FIXME: This is a workaround to load the full DF before saving it. The
        # DF is then reloaded, loading only the units just wrangled. This
        # avoids using memory but needs cleanup.
        #
        store_df(units, etype='unit')
        stored_units = load_df(etype='unit', sid=sid)
        print('\nstored in %s.' % csvs['unit'])

    # Returns the DF of the stored pilots if no new pilots have been added;
    # the DF with the old and new pilots otherwise.
    return stored_units


# ------------------------------------------------------------------------------
#
def load_session(sid, exp, sra_session, sra_pilots, sra_units,
                 sdm, pdm, udm, pilots, units, sts):

    # REDUNDANT: get_new_sessions checks for this already
    # If this session has been already stored get out, nothing to do here.
    # stored_sessions = load_df(etype='session', sid=sid)
    # if sid in stored_sessions.index.tolist():
    #     sys.stdout.write('%s already stored in %s' % (sid, csvs['session']))
    #     return False

    sys.stdout.write('\n%s --- %s' % (exp, sid))
    s = initialize_entity(etype='session')

    # Session properties: pilots and units.
    # sp = sra_session.filter(etype='pilot', inplace=False)
    # su = sra_session.filter(etype='unit', inplace=False)
    s['sid'].append(sid)
    s['session'].append(None)
    s['experiment'].append(exp)
    s['nhost'].append(len(pilots.loc[pilots['sid'] == sid]['hid'].unique()))
    s['nunit'].append(len(sra_units.get()))
    s['npilot'].append(len(sra_pilots.get()))
    s['npilot_active'].append(len(sra_pilots.timestamps(state='PMGR_ACTIVE')))
    s['nunit_done'].append(len(sra_units.timestamps(state='DONE')))
    s['nunit_failed'].append(len(sra_units.timestamps(state='FAILED')))

    # Number of cores requested and used by the session's pilots. Make a copy of
    # the pilots DF with only the columns we need to limit memory overhead.
    pcores = pilots[pilots.sid == sid][['P_LRMS_RUNNING', 'ncore']]
    s['ncore'].append(pcores.ncore.sum())
    s['ncore_active'].append(pcores[pcores.P_LRMS_RUNNING > 0].ncore.sum())
    pcores = None

    # Session timestamps.
    # for state in sts:
    #     if state not in s:
    #         s[state] = list()
        # FIXME: Open a ticket requesting timestamps for session's events.
        # s[state].append(min(sra_pilots.timestamps(state=state)))
    s['NEW'] = list()
    s['NEW'].append(min(sra_pilots.timestamps(state='NEW')))

    # Session total duration.
    # for duration in sdm:
    #     if duration not in s:
    #         s[duration] = list()
    #     s['TTC'].append(sra_session.ttc)
    s['TTC'] = list()
    s['TTC'].append(sra_session.ttc)

    # Pilots total durations.  NOTE: s initialization guarantees
    # the existence of duration keys.
    for duration in pdm:
        s[duration].append(sra_pilots.duration(event=pdm[duration]))

    # Units total durations. NOTE: s initialization guarantees the
    # existence of duration keys.
    for duration in udm:
        s[duration].append(sra_units.duration(event=udm[duration]))

    # Store session.
    session = pd.DataFrame(s, index=[sid])
    # FIXME: This is a workaround to load the full DF before saving it.
    stored_sessions = load_df(etype='session')
    store_df(session, stored=stored_sessions, etype='session')
    print('\nstored in %s' % csvs['session'])

    return stored_sessions


# ------------------------------------------------------------------------------
#
def get_raw_sessions(ddir, etag, clopts):
    '''
    Get sessions ID, experiment number and RA object.
    Assume: ddir/exp*/sessiondir/session.json.
    '''

    print('\n\nGet raw sessions: ')
    sids = dict()

    for path in glob.glob('%s/%s*' % (ddir, etag)):

        # Every directory in the experiment dir is assumed to be a RP session.
        for sdir in glob.glob('%s/*/' % path):

            # we assume the dir name is the session ID (after stripping '/'-es)
            sdir = sdir.rstrip('/')
            sid  = os.path.basename(sdir)

            # Skip session if not specified at command line.
            if clopts['sid'] and clopts['sid'] != sid:
                print('Ignore directory %s; waiting for %s' % (sdir,
                    clopts['sid']))
                continue

            # Get session ID directory if json file exists.
            # sid = glob.glob('%s/*.json' % sdir)[0].split('/')[-2]
            if not os.path.isfile('%s/%s.json' % (sdir, sid)):
                print("ERROR: no session json in %s" % sdir)
                print("Skip %s" % sdir)
                continue

            # Get experiment directory.
            exp = os.path.basename(path)

            # Skip session if not in the experiment directory specified at
            # command line.
            # FIXME: AM: I don't understand the index notation below
            if clopts['eid'] and clopts['eid'] != exp[len(etag):]:
                print('Ignore %s; looking for directories in %s' % (
                    exp[len(etag):], clopts['eid']))
                continue

            # all tests passed - we want this session
            sids[sdir] = sid

    print('Done.')
    return sids


# ------------------------------------------------------------------------------
#
def get_new_sessions(sids):
    '''
    For all sid's given, check if any needs wrangling, and return a dict with
    those session IDs.
    '''

    print('\n\nMarking sessions for wrangling: ')
    towrangle = dict()

    # Load current sessions, pilots, units DFs
    sessions = load_df(etype='session')
    pilots   = load_df(etype='pilot')
    units    = load_df(etype='unit')

    # All sessions are new if we have no stored sessions.
    if sessions.empty:
        return sids

    # Add a session to wrangler when the sesison is not in the sessions DF or
    # when any of the session's units or pilots are not in the units or pilots
    # DFs.
    for sdir, sid in sids.items():
        nurequest = sessions[sessions.sid == sid].nunit.tolist()
        nprequest = sessions[sessions.sid == sid].npilot.tolist()

        # Duplicates in sessions need to be addressed manually by the user.
        # TODO: AM: ?
        if (len(nurequest) > 1) or (len(nprequest) > 1):
            print('ERROR: Duplicate entries for sid %s in sessions df' % sid)
            sys.exit(1)

        if  sid not in sessions.sid.tolist()                  or \
            units[units.sid   == sid].shape[0] < nurequest[0] or \
            pilots[pilots.sid == sid].shape[0] < nprequest[0] :
            print('Mark session %s for wrangling' % sid)
            towrangle[sdir] = sid

    print('Done.')
    return towrangle


# ------------------------------------------------------------------------------
#
def wrangle_session(sdir, sid):

    # Get the experiment tag for the current sdir.
    exp = sdir.split('/')[-2:][0]

    # RA objects cannot be serialize: every RA session object need
    # to be constructed at every run.
    #
    # FIXME: AM: Actually, RA sessions can be pickled all right:
    #
    #        import radical.analytics as ra
    #        import pickle
    #        s1 = ra.Session(src=sdir, stype='radical.pilot')
    #        p  = pickle.dumps(s1)
    #        s2 = pickle.loads(p)
    #        assert(len(s1.get()) == len(s2.get()))
    #
    #        If we found cases where this does not work, we can make it work.
    #
    sra_session = ra.Session(sdir, 'radical.pilot')

    # Pilot-unit relationship dictionary
    pu_rels = sra_session.describe('relations', ['pilot', 'unit'])

    # Pilots of sra: dervie properties and durations.
    print('\n\n%s -- %s -- Loading pilots:' % (exp, sid))
    sra_pilots = sra_session.filter(etype='pilot', inplace=False)
    pilots     = load_pilots(sid, exp, sra_pilots, pdm, pu_rels, pts)

    # Units of sra: dervie properties and durations.
    print('\n\n%s -- %s -- Loading units:' % (exp, sid))
    sra_units = sra_session.filter(etype='unit', inplace=False)
    units     = load_units(sid, exp, sra_units, udm, pilots,
                           sra_session, pu_rels, uts)

    # Session of sra: derive properties and total durations.
    print('\n\n%s -- %s -- Loading session:\n' % (exp, sid))
    load_session(sid, exp, sra_session, sra_pilots, sra_units,
                 sdm, pdm, udm, pilots, units, sts)


# =============================================================================
if __name__ == '__main__':

    # Get command line options
    clopts = clparse(sys.argv[1:])

    # Where to find data (ddir) and how data are stored into experiments
    # (etag).
    ddir = clopts['ddir']  # e.g., '../data/'
    etag = clopts['etag']  # e.g., 'exp'

    # File names where to save the DF of each entity of each session.
    csvs = {'session': '%s/sessions.csv' % clopts['odir'],
            'pilot'  : '%s/pilots.csv'   % clopts['odir'],
            'unit'   : '%s/units.csv'    % clopts['odir']}

    # FIXME: Define timestamps of the events of the pilot's states.
    sts = {rp.NEW     : None,
           rp.DONE    : None,
           rp.CANCELED: None,
           rp.FAILED  : None}

    # FIXME: Define session durations.
    sdm = {'TTC': None}

    # Define timestamps of the events of the pilot's states.
    pts = {rp.NEW                   : None,
           rp.PMGR_LAUNCHING_PENDING: None,
           rp.PMGR_LAUNCHING        : None,
           rp.PMGR_ACTIVE_PENDING   : None,
           rp.PMGR_ACTIVE           : None,
           rp.DONE                  : None,
           rp.CANCELED              : None,
           rp.FAILED                : None}

    # Define pilot durations.
    pdm = {'P_PMGR_SCHEDULING': [{ru.EVENT: 'state',             ru.STATE: rp.NEW                   },
                                 {ru.EVENT: 'state',             ru.STATE: rp.PMGR_LAUNCHING_PENDING}],
           'P_PMGR_QUEUING'   : [{ru.EVENT: 'state',             ru.STATE: rp.PMGR_LAUNCHING_PENDING},
                                 {ru.EVENT: 'state',             ru.STATE: rp.PMGR_LAUNCHING        }],
           'P_LRMS_SUBMITTING': [{ru.EVENT: 'state',             ru.STATE: rp.PMGR_LAUNCHING        },
                                 {ru.EVENT: 'state',             ru.STATE: rp.PMGR_ACTIVE_PENDING   }],
           'P_LRMS_QUEUING'   : [{ru.EVENT: 'state',             ru.STATE: rp.PMGR_ACTIVE_PENDING   },
                                 {ru.EVENT: 'state',             ru.STATE: rp.PMGR_ACTIVE           }],
           'P_LRMS_RUNNING'   : [{ru.EVENT: 'state',             ru.STATE: rp.PMGR_ACTIVE           },
                                [{ru.EVENT: 'state',             ru.STATE: rp.DONE                  },
                                 {ru.EVENT: 'state',             ru.STATE: rp.FAILED                },
                                 {ru.EVENT: 'state',             ru.STATE: rp.CANCELED              }]],

           # from utilization script
           'util_p_total'     : [{ru.EVENT: 'bootstrap_1_start', ru.STATE: None                     },
                                 {ru.EVENT: 'bootstrap_1_stop',  ru.STATE: None                     }],
           'util_p_boot'      : [{ru.EVENT: 'bootstrap_1_start', ru.STATE: None                     },
                                 {ru.EVENT: 'sync_rel',          ru.STATE: None                     }],
           'util_p_setup_1'   : [{ru.EVENT: 'sync_rel',          ru.STATE: None                     },
                                 {ru.EVENT: 'orte_dvm_start',    ru.STATE: None                     }],
           'util_p_orte'      : [{ru.EVENT: 'orte_dvm_start',    ru.STATE: None                     },
                                 {ru.EVENT: 'orte_dvm_ok',       ru.STATE: None                     }],
           'util_p_setup_2'   : [{ru.EVENT: 'orte_dvm_ok',       ru.STATE: None                     },
                                 {ru.EVENT: 'state',             ru.STATE: rp.PMGR_ACTIVE           }],
           'util_p_uexec'     : [{ru.EVENT: 'state',             ru.STATE: rp.PMGR_ACTIVE           },
                                 {ru.EVENT: 'cmd',               ru.STATE: None                     }],
           'util_p_term'      : [{ru.EVENT: 'cmd',               ru.STATE: None                     },
                                 {ru.EVENT: 'bootstrap_1_stop',  ru.STATE: None                     }],
          }

    # Define timestamps of the events of the pilot's states.
    uts = {rp.NEW                         : None,
           rp.UMGR_SCHEDULING_PENDING     : None,
           rp.UMGR_SCHEDULING             : None,
           rp.UMGR_STAGING_INPUT_PENDING  : None,
           rp.UMGR_STAGING_INPUT          : None,
           rp.AGENT_STAGING_INPUT_PENDING : None,
           rp.AGENT_STAGING_INPUT         : None,
           rp.AGENT_SCHEDULING_PENDING    : None,
           rp.AGENT_SCHEDULING            : None,
           rp.AGENT_EXECUTING_PENDING     : None,
           rp.AGENT_EXECUTING             : None,
           rp.AGENT_STAGING_OUTPUT_PENDING: None,
           rp.AGENT_STAGING_OUTPUT        : None,
           rp.UMGR_STAGING_OUTPUT_PENDING : None,
           rp.UMGR_STAGING_OUTPUT         : None,
           rp.DONE                        : None,
           rp.CANCELED                    : None,
           rp.FAILED                      : None}

    # Define unit durations.
    udm = {'U_UMGR_SCHEDULING'      : [{ru.EVENT: 'state',                  ru.STATE: rp.NEW                         },
                                       {ru.EVENT: 'state',                  ru.STATE: rp.UMGR_SCHEDULING_PENDING     }],
           'U_UMGR_BINDING'         : [{ru.EVENT: 'state',                  ru.STATE: rp.UMGR_SCHEDULING_PENDING     },
                                       {ru.EVENT: 'state',                  ru.STATE: rp.UMGR_SCHEDULING             }],
         # 'I_UMGR_SCHEDULING'      : [{ru.EVENT: 'state',                  ru.STATE: rp.UMGR_SCHEDULING             },
         #                             {ru.EVENT: 'state',                  ru.STATE: rp.UMGR_STAGING_INPUT_PENDING  }],
         # 'I_UMGR_QUEING'          : [{ru.EVENT: 'state',                  ru.STATE: rp.UMGR_STAGING_INPUT_PENDING  },
         #                             {ru.EVENT: 'state',                  ru.STATE: rp.UMGR_STAGING_INPUT          }],
         # 'I_AGENT_SCHEDULING'     : [{ru.EVENT: 'state',                  ru.STATE: rp.UMGR_STAGING_INPUT          },
         #                             {ru.EVENT: 'state',                  ru.STATE: rp.AGENT_STAGING_INPUT_PENDING }],
         # 'I_AGENT_QUEUING'        : [{ru.EVENT: 'state',                  ru.STATE: rp.AGENT_STAGING_INPUT_PENDING },
         #                             {ru.EVENT: 'state',                  ru.STATE: rp.AGENT_STAGING_INPUT         }],
         # 'I_AGENT_TRANSFERRING'   : [{ru.EVENT: 'state',                  ru.STATE: rp.AGENT_STAGING_INPUT         },
         #                             {ru.EVENT: 'state',                  ru.STATE: rp.AGENT_SCHEDULING_PENDING    }],
           'U_AGENT_QUEUING'        : [{ru.EVENT: 'state',                  ru.STATE: rp.AGENT_SCHEDULING_PENDING    },
                                       {ru.EVENT: 'state',                  ru.STATE: rp.AGENT_SCHEDULING            }],
           'U_AGENT_SCHEDULING'     : [{ru.EVENT: 'state',                  ru.STATE: rp.AGENT_SCHEDULING            },
                                       {ru.EVENT: 'state',                  ru.STATE: rp.AGENT_EXECUTING_PENDING     }],
           'U_AGENT_QUEUING_EXEC'   : [{ru.EVENT: 'state',                  ru.STATE: rp.AGENT_EXECUTING_PENDING     },
                                       {ru.EVENT: 'state',                  ru.STATE: rp.AGENT_EXECUTING             }],
           'U_AGENT_EXECUTING'      : [{ru.EVENT: 'state',                  ru.STATE: rp.AGENT_EXECUTING             },
                                       {ru.EVENT: 'state',                  ru.STATE: rp.AGENT_STAGING_OUTPUT_PENDING}],
         # 'O_AGENT_QUEUING'        : [{ru.EVENT: 'state',                  ru.STATE: rp.AGENT_STAGING_OUTPUT_PENDING},
         #                             {ru.EVENT: 'state',                  ru.STATE: rp.AGENT_STAGING_OUTPUT        }],
         # 'O_UMGR_SCHEDULING'      : [{ru.EVENT: 'state',                  ru.STATE: rp.AGENT_STAGING_OUTPUT        },
         #                             {ru.EVENT: 'state',                  ru.STATE: rp.UMGR_STAGING_OUTPUT_PENDING }],
         # 'O_UMGR_QUEUING'         : [{ru.EVENT: 'state',                  ru.STATE: rp.UMGR_STAGING_OUTPUT_PENDING },
         #                             {ru.EVENT: 'state',                  ru.STATE: rp.UMGR_STAGING_OUTPUT         }],
         # 'O_UMGR_TRANSFERRING'    : [{ru.EVENT: 'state',                  ru.STATE: rp.UMGR_STAGING_OUTPUT         },
         #                            [{ru.EVENT: 'state',                  ru.STATE: rp.DONE                        },
         #                             {ru.EVENT: 'state',                  ru.STATE: rp.FAILED                      },
         #                             {ru.EVENT: 'state',                  ru.STATE: rp.CANCELED                    }]],

           # from utilization script
            'util_u_total'          : [{ru.EVENT: 'schedule_ok',            ru.STATE: None                           },
                                       {ru.EVENT: 'unschedule_stop',        ru.STATE: None                           }],
            'cu_exec_start_stop'    : [{ru.EVENT: 'cu_exec_start',          ru.STATE: None                           },
                                       {ru.EVENT: 'cu_exec_stop',           ru.STATE: None                           }],
            'util_u_equeue'         : [{ru.EVENT: 'schedule_ok',            ru.STATE: None                           },
                                       {ru.EVENT: 'state',                  ru.STATE: rp.AGENT_EXECUTING             }],
            'util_u_pre_exec'       : [{ru.EVENT: 'cu_pre_start',           ru.STATE: None                           },
                                       {ru.EVENT: 'cu_pre_stop',            ru.STATE: None                           }],
            'util_u_eprep'          : [{ru.EVENT: 'state',                  ru.STATE: rp.AGENT_EXECUTING             },
                                       {ru.EVENT: 'exec_start',             ru.STATE: None                           }],
            'util_u_exec_rp'        : [{ru.EVENT: 'exec_start',             ru.STATE: None                           },
                                       {ru.EVENT: 'cu_start',               ru.STATE: None                           }],
            'util_u_exec_cu'        : [{ru.EVENT: 'cu_start',               ru.STATE: None                           },
                                       {ru.EVENT: 'cu_exec_start',          ru.STATE: None                           }],
            'util_u_exec_orte'      : [{ru.EVENT: 'cu_exec_start',          ru.STATE: None                           },
                                       {ru.EVENT: 'app_start',              ru.STATE: None                           }],
            'util_u_exec_app'       : [{ru.EVENT: 'app_start',              ru.STATE: None                           },
                                       {ru.EVENT: 'app_stop',               ru.STATE: None                           }],
            'util_u_unschedule'     : [{ru.EVENT: 'app_stop',               ru.STATE: None                           },
                                       {ru.EVENT: 'unschedule_stop',        ru.STATE: None                           }],
          }


    # Find out what sessions need to be wrangled.
    rawsids = get_raw_sessions(ddir, etag, clopts)
    sids    = get_new_sessions(rawsids)
    procs   = list()

    print(sids) 

    num_workers = psutil.cpu_count(logical=False)
    workers     = mp.Pool(num_workers)

    # Wrangle the new sessions.
    if sids:
        for sdir,sid in sids.items():
            wrangle_session(sdir, sid)
        #     workers.apply_async(wrangle_session, (sdir, sid,))
        # workers.close()
        # workers.join()

    else:
        print('No new sessions to wrangle found.')

# ------------------------------------------------------------------------------




#!/usr/bin/env python


import sys
import time
import optparse

import numpy             as np
import matplotlib.pyplot as plt
import radical.analytics as ra

from radical.analytics.utils import to_latex


# ----------------------------------------------------------------------------
#
plt.style.use(ra.get_mplstyle("radical_mpl"))


# ------------------------------------------------------------------------------
#
TITLE     = ''
DELIM     = ''
MATCH     = None
COLUMN_X  = 'count'
COLUMNS_Y = ['1']
LEGEND    = []
LABEL_X   = ''
LABEL_Y   = ''
TICKS_X   = []
TICKS_Y   = []
RANGE     = [None, None, None, None]
LOG_X     = False
LOG_Y     = False
LOG       = ''
STYLE     = 'line'  # 'point', 'line', 'step', 'bar', 'hist', 'lp'
GRID      = False   # True, False
FNAME     = None
SAVE_AS   = 'x11'   # 'svg', 'png', 'x11', 'pdf'
WIDTH     = 500
HEIGHT    = None
STDEV     = False


# ------------------------------------------------------------------------------
#
def usage(msg=None):

    ret = 0
    if msg:
        print('\n\n\t\tError: %s\n' % msg)
        ret = 1

    print('''

    usage  : radical-analytics-plot [<src>] [options]
    example: radical-analytics-plot experiment.dat -t 'title' --log 'x,y'
    source : space separated data file (read from stdin if not specified)
    options:
        -h, --help                             : this message
        -t, --title      <title>               : plot title
        -X, --x-label    <label>               : x-axis label
        -Y, --y-label    <label>               : y-axis label
        -d, --delim      <char>                : column delimiter
        -m, --match      <pattern>             : use lines matching pattern
        -x, --x-column   <col>                 : source column for x-values
        -y, --y-columns  <col_1,col_2,...>     : list of columns columns to plot
        -S, --stdev                            : plot stdev from `col + 1`
        -L, --legend     <label_1,label_2,...> : name of plots specified in '-y'
        -u, --x-ticks    <tick_1,tick_2,...>   : not yet supported
        -v, --y-ticks    <tick_1,tick_2,...>   : not yet supported
        -r, --range      <xmin,xmax,ymin,ymax> : axis range
        -s, --style      <point | line | step | bar | hist | lp>
                                               : plot type
        -W, --width      <pixels>              : canvas width (default: 500)
        -H, --height     <pixels>              : canvas height
        -l, --log        <x | y | x,y>         : log-scale for x and/or y axis
        -g, --grid                             : grid lines (default: no)
        -a, --save-as    <png | svg | x11>     : save fig in format (x11: show)
        -f, --file-name  <filename>            : name to save to (w/o ext)

''')

    sys.exit(ret)


# ------------------------------------------------------------------------------
# parse options
parser = optparse.OptionParser(add_help_option=False)

parser.add_option('-t', '--title',     dest='title')
parser.add_option('-X', '--x-label',   dest='xlabel')
parser.add_option('-Y', '--y-label',   dest='ylabel')
parser.add_option('-d', '--delimiter', dest='delim')
parser.add_option('-m', '--match',     dest='match')
parser.add_option('-x', '--x-column',  dest='xcol')
parser.add_option('-y', '--y-columns', dest='ycols')
parser.add_option('-u', '--x-ticks',   dest='xticks')
parser.add_option('-v', '--y-ticks',   dest='yticks')
parser.add_option('-S', '--stdev',     dest='stdev', action='store_true')
parser.add_option('-r', '--range',     dest='range')
parser.add_option('-L', '--legend',    dest='legend')
parser.add_option('-s', '--style',     dest='style')
parser.add_option('-W', '--width',     dest='width')
parser.add_option('-H', '--height',    dest='height')
parser.add_option('-l', '--log',       dest='log')
parser.add_option('-g', '--grid',      dest='grid', action='store_true')
parser.add_option('-a', '--save-as',   dest='save')
parser.add_option('-f', '--file-name', dest='fname')
parser.add_option('-h', '--help',      dest='help', action='store_true')

options, args = parser.parse_args()
if len(args) > 1:
    usage("Too many arguments (%s)" % args)

if len(args) < 1:
    src = None
else:
    src = args[0]

if options.help   : usage()
if options.title  : TITLE        =  str(options.title)
if options.delim  : DELIM        =  str(options.delim)
if options.match  : MATCH        =  str(options.match)
if options.xcol   : COLUMN_X     =  str(options.xcol)
if options.ycols  : COLUMNS_Y    = [str(x) for x in options.ycols .split(',')]
if options.xlabel : LABEL_X      =  str(options.xlabel)
if options.ylabel : LABEL_Y      =  str(options.ylabel)
if options.xticks : TICKS_X      = [str(x) for x in options.xticks.split(',')]
if options.yticks : TICKS_Y      = [str(x) for x in options.yticks.split(',')]
if options.legend : LEGEND       = [str(x) for x in options.legend.split(',')]
if options.width  : WIDTH        =  int(options.width)
if options.height : HEIGHT       =  int(options.height)
if options.log    : LOG          =  str(options.log)
if options.style  : STYLE        =  str(options.style)
if options.save   : SAVE_AS      =  str(options.save)
if options.fname  : FNAME        =  str(options.fname)
if options.stdev  : STDEV        =  True
if options.grid   : GRID         =  True

if options.range  :
    RANGE = options.range .split(',')
    RANGE = [float(x) if x else None for x in RANGE]

LEGEND  = [s.strip() for s in LEGEND]
TICKS_X = [s.strip() for s in TICKS_X]
TICKS_Y = [s.strip() for s in TICKS_Y]

if 'x' in LOG: LOG_X = True
if 'y' in LOG: LOG_Y = True

if COLUMN_X not in ['count']:
    COLUMN_X = int(COLUMN_X)

if SAVE_AS not in ['x11', 'png', 'svg', 'pdf']:
    raise ValueError('invalid save_as value: %s' % SAVE_AS)

if STYLE not in ['point', 'line', 'step', 'bar', 'hist', 'lp']:
    raise ValueError('invalid style: %s' % STYLE)


# ------------------------------------------------------------------------------
# read and sort data
def get_lines():
    if src:
        with open(src, 'r') as fin:
            for line in fin.readlines():
                line = line.strip()
                if line:
                    yield line
    else:
        for line in sys.stdin:
            line = line.strip()
            if line:
                yield line


# ------------------------------------------------------------------------------
def get_elems(line):
    if DELIM:
        elems = [e.strip() for e in line.split(DELIM)]
    else:
        elems = line.split()

    ret = list()
    for e in elems:
        try:
            ret.append(float(e))
        except:
            ret.append(e)

    return ret


# ------------------------------------------------------------------------------
rows = list()
for line in get_lines():

    if line.startswith('#'):
      # continue
        elems  = get_elems(line)[1:]
        if not LEGEND:
            LEGEND = [''] * (len(elems))
        for idx in range(len(elems)):
            elem = elems[idx]
            if idx == COLUMN_X:
                if LABEL_X == '-':
                    LABEL_X = None
                else:
                    LABEL_X = elem
            else:
                if not LEGEND:
                    LEGEND[idx] = elem

    else:
        if MATCH and MATCH not in line:
            continue
        rows.append(get_elems(line))

    if not LABEL_X:
        LABEL_X = COLUMN_X

if not rows:
    raise ValueError('no matching data')


# ------------------------------------------------------------------------------
# invert rows to actual data layout
data  = list()
ncols = len(rows[0])
for idx in range(ncols):
    data.append(list())

for row in rows:
    for col in range(ncols):
        data[col].append(row[col])

if STYLE == 'hist':
    if COLUMN_X and COLUMN_X != 'count':
        raise ValueError('histogram plots should specify `-y`, not `-x`')

# ------------------------------------------------------------------------------
# plot data
# pprint.pprint(data)
try:

    fig, ax = plt.subplots(figsize=ra.get_plotsize(width=WIDTH, height=HEIGHT))
    cnum = 0
    for col in COLUMNS_Y:

        data_y_err = None

        if   LEGEND[0] == ['-']: label = None,
        elif LEGEND            : label = to_latex(LEGEND[cnum])
        else                   : label = to_latex(str(cnum))

        cnum += 1

        if COLUMN_X == 'count':
            data_x = list(range(len(data[0])))
        else:
            data_x = np.array(data[COLUMN_X])


        if '+' in col:
            cols   = [int(c) for c in col.split('+')]
            data_y = sum([np.array(data[c]) for c in cols])
            if STDEV:
                data_y_err = sum([np.array(data[c + 1]) for c in cols])

        elif '-' in col:
            cols   = [int(c) for c in col.split('-')]
            data_y = np.array(data[cols[0]])
            if STDEV:
                data_y_err = np.array(data[cols[0] + 1])
            for c in cols[1:]:
                data_y -= np.array(data[c])
                if STDEV:
                    data_y_err += np.array(data[c + 1])

        elif '*' in col:
            cols   = [int(c) for c in col.split('*')]
            data_y = np.array(data[cols[0]])
            if STDEV:
                data_y_err = np.array(data[cols[0] + 1])
            for c in cols[1:]:
                data_y *= np.array(data[c])
                if STDEV:
                    data_y_err += np.array(data[c + 1])

        elif '/' in col:
            cols   = [int(c) for c in col.split('/')]
            data_y = np.array(data[cols[0]])
            if STDEV:
                data_y_err = np.array(data[cols[0] + 1])
            for c in cols[1:]:
                data_y /= np.array(data[c])
                if STDEV:
                    data_y_err += np.array(data[c + 1])

        else:
            col = int(col)
            time.sleep(1)
            data_y  = np.array(data[col])
            if STDEV:
                data_y_err = np.array(data[col + 1])

        if STDEV and data_y_err is None:
            raise ValueError('stdev not available with column spec %s' % col)

        colors = plt.rcParams['axes.prop_cycle'].by_key()['color']
        color = colors[cnum]

        # if cnum == 3:
        #     color = colors[4]
        if   STYLE == 'point': ax.scatter(data_x, data_y, c=color, label=label, s=10)
        elif STYLE == 'line' : ax.plot   (data_x, data_y, c=color, label=label)
        elif STYLE == 'step' : ax.step   (data_x, data_y, c=color, label=label)
        elif STYLE == 'bar'  : ax.bar    (data_x, data_y, c=color, label=label)
        elif STYLE == 'hist' : ax.hist   (data_y,  150,   color=color, label=label)
        elif STYLE == 'lp'   : ax.plot   (data_x, data_y, c=color, label=label, marker='.')

      # if STDEV:
      #     ax.fill_between(data_x, data_y - data_y_err, data_y + data_y_err,
      #                     color=color, alpha=0.1)

        if STDEV:
            ax.errorbar(data_x, data_y, yerr=data_y_err, fmt='.', color=color)

except IndexError:
    print('index error')
    for i,e in enumerate(data[0]):
        print('    %2d: %s' % (i, e))
    raise

print(LEGEND)
if LEGEND[0] != '-':
    plt.legend(ncol=len(LEGEND), bbox_to_anchor=(1.05, 1.25),
          fancybox=True)

# ax.ytick.label_format(axis='both', style='sci', scilimits=(2,None))

if TITLE   : ax.set_title(TITLE, loc='center')
if LOG_X   : ax.set_xscale('log')
if LOG_Y   : ax.set_yscale('log')
if LABEL_X : ax.set_xlabel(to_latex(LABEL_X))
if LABEL_Y : ax.set_ylabel(to_latex(LABEL_Y))
if TICKS_X : ax.set_xticks([int(t) for t in TICKS_X], TICKS_X)
if TICKS_Y : ax.set_yticks([int(t) for t in TICKS_Y], TICKS_Y, )
if GRID    : ax.grid(True)

xmin, xmax, ymin, ymax = RANGE
if xmin is not None: ax.set_xlim(left=xmin)
if xmax is not None: ax.set_xlim(right=xmax)
if ymin is not None: ax.set_ylim(bottom=ymin)
if ymax is not None: ax.set_ylim(top=ymax)

if not FNAME:
    FNAME = TITLE.lower().replace(' ', '_')
    FNAME = FNAME.replace(':', '_-_')
    FNAME = FNAME.replace('__', '_')

if   SAVE_AS == 'png': fig.savefig('%s.png' % FNAME, bbox_inches="tight")
elif SAVE_AS == 'svg': fig.savefig('%s.svg' % FNAME, bbox_inches="tight")
elif SAVE_AS == 'pdf': fig.savefig('%s.pdf' % FNAME, dpi=300,
                                              bbox_inches="tight")
elif SAVE_AS == 'x11': fig.show()


# ------------------------------------------------------------------------------





# import radical.analytics as ra
import radical.pilot as rp
# import os
# import glob


# ------------------------------------------------------------------------------
#
def get_duration_ra(session):

    '''
    Get duration using the 'duration' method of the session object in analytics
    '''
    units = session.filter(etype='unit', inplace=False)
    return units.duration([rp.UMGR_SCHEDULING_PENDING, rp.DONE])


# ------------------------------------------------------------------------------
#
def get_duration_minmax(session):

    '''
    Get max of the final state and min of the initial state for all units
    '''

    # FIXME: this is a useless test - semantically it tests if
    #        ru.collapse_ranges maintains time ordering, which should be tested
    #        in RU, not here.  Also, it will fail for partitioned ranges, so
    #        will not work for all test cases.

    units_1  = session.get(state=rp.UMGR_SCHEDULING_PENDING, etype='unit')
    start_rp = [unit.states[rp.UMGR_SCHEDULING_PENDING]['time']
                            for unit in units_1]

    units_2  = session.get(state=rp.DONE, etype='unit')
    stop_rp = [unit.states[rp.DONE]['time'] for unit in units_2]

    return max(stop_rp) - min(start_rp)


# ------------------------------------------------------------------------------
#
def collapse_ranges(ranges):
    """
    given be a set of ranges (as a set of pairs of floats [start, end] with
    'start <= end'. This algorithm will then collapse that set into the
    smallest possible set of ranges which cover the same, but not more nor
    less, of the domain (floats).

    We first sort the ranges by their starting point. We then start with the
    range with the smallest starting point [start_1, end_1], and compare to
    the next following range [start_2, end_2], where we now know that
    start_1 <= start_2. We have now two cases:

    a) when start_2 <= end_1, then the ranges overlap, and we collapse them
    into range_1: range_1 = [start_1, max[end_1, end_2]

    b) when start_2 > end_2, then ranges don't overlap. Importantly, none of
    the other later ranges can ever overlap range_1. So we move range_1 to
    the set of final ranges, and restart the algorithm with range_2 being
    the smallest one.

    Termination condition is if only one range is left -- it is also moved
    to the list of final ranges then, and that list is returned.
    """

    # FIXME: this is a useless helper: it implements ru.collapse_ranges
    #        again, and of course the same result appears - duh! We can
    #        rely external methods to be tested.

    final = []

    # sort ranges into a copy list
    _ranges = sorted (ranges, key=lambda x: x[0])

    START = 0
    END = 1

    base = _ranges[0]  # smallest range

    for _range in _ranges[1:]:

        if _range[START] <= base[END]:

            # ranges overlap -- extend the base
            base[END] = max(base[END], _range[END])

        else:

            # ranges don't overlap -- move base to final, and current _range
            # becomes the new base
            final.append(base)
            base = _range

    # termination: push last base to final
    final.append(base)

    return final


# ------------------------------------------------------------------------------
#
def get_duration_ru(session):

    '''
    Use the sliding window method in radical.utils to derive the duration
    between the two states for all units
    '''

    # FIXME: This is a useless test: it implements ra.session.duration again and
    #        checks if it jields the same result.  Duh, what else could happen?

    units  = session.get(state=[rp.UMGR_SCHEDULING_PENDING, rp.DONE],
                         etype='unit')

    ranges = [[unit.states[rp.UMGR_SCHEDULING_PENDING]['time'],
               unit.states[rp.DONE                   ]['time']]
              for unit in units]

    overlap = 0.0
    for crange in collapse_ranges(ranges):
        overlap += crange[1] - crange[0]

    return overlap


# ------------------------------------------------------------------------------
#
def test_duration_no_barrier():

    '''
    This function tests if the durations obtained from the analytics function is
    the same as the duration obtained by the utils function. They should both be
    equal to the 'max-min' (of the FINAL and INITIAL states respectively) as the
    data set in this case consists of profiles when all units are concurrently
    being executed, i.e. there is no execution barrier between them and thus no
    'gap'/full overlap.
    '''

    # FIXME: data files broken
    return

    # data_loc   = '%s/no_barrier_data' % os.path.dirname(__file__)
    # json_files = glob.glob('%s/*.json' % data_loc)
    # json_file  = json_files[0]
    # sid        = os.path.basename(json_file)[:-5]

    # session = ra.Session(sid=sid, stype='radical.pilot', src=data_loc)

    # assert get_duration_ra(session) == get_duration_ru(session)
    # assert get_duration_ra(session) == get_duration_minmax(session)


# ------------------------------------------------------------------------------
#
def test_duration_method_barrier():

    '''
    This function tests if the durations obtained from the analytics function is
    the same as the duration obtained by the utils function. They should both be
    less than the 'max-min' (of the FINAL and INITIAL states respectively) as
    the data set in this case consists of profiles when not all units are
    concurrently being executed, i.e. there is an execution barrier between them
    and thus a 'gap' between their executions.
    '''

    # FIXME: data files broken
    return

    # data_loc   = '%s/barrier_data' % os.path.dirname(__file__)
    # json_files = glob.glob('%s/*.json' % data_loc)
    # json_file  = json_files[0]
    # sid        = os.path.basename(json_file)[:-5]

    # session = ra.Session(sid=sid, stype='radical.pilot', src=data_loc)

    # assert get_duration_ra(session) == get_duration_ru(session)
    # assert get_duration_ra(session) <  get_duration_minmax(session)


# ------------------------------------------------------------------------------




#!/usr/bin/env python

__copyright__ = 'Copyright 2013-2016, http://radical.rutgers.edu'
__license__   = 'MIT'


import sys

import matplotlib.pyplot as plt

import radical.utils     as ru
import radical.analytics as ra

from radical.analytics.utils import to_latex


# ----------------------------------------------------------------------------
#
plt.style.use(ra.get_mplstyle("radical_mpl"))



# This script plots the core utilization of a set of RP sessions in a stacked
# barplot, where the stacked elements represent how the cores were (or were not)
# utilized.  The elements will always add up to 100%, thus the whole bar
# represents the amount of core-hours available to the session.  We only look at
# events on the agent side.
#
# The bar elements are determined as follows (for each pilot)
#
#  - During bootstrap, configuration, setup, and termination, cores are
#    essentially unused.  We separate out the times needed for those steps, and
#    multiply by pilot size, to derive the number of core hours essentially
#    spent on those activities (see `PILOT_DURATIONS` below).
#
#    NOTE: we assume that the pilot activities thus measured stop at the point
#          when the first task gets scheduled on a (set of) core(s), and
#          restarts when the last task gets unscheduled, as that is the time
#          frame where we in principle consider cores to be available to the
#          workload.
#
#  - For each task, we look at the amount of time that task has been scheduled
#    on a set of cores, as those cores are then essentially blocked.  Multiplied
#    by the size of the task, that gives a number of core-hours those cores are
#    'used' for that task.
#
#    not all of that time is utilized for application use though: some is spent
#    on preparation for execution, on spawning, unscheduling etc.  We separate
#    out those utilizations for each task.
#
#  - we consider core hours to be additive, in the following sense:
#
#    - all core-hours used by the pilot in various global activities listed in
#      the first point, plus the sumof core hours spend by all tasks in various
#      individual activities as in the second point, equals the overall core
#      hours available to the pilot.
#
#      This only holds with one caveat: after the agent started to work on task
#      execution, some cores may not *yet* be allocated (scheduler is too slow),
#      or may not be allocated *anymore* (some tasks finished, we wait for the
#      remaining ones).  We consider those core-hours as 'idle'.
#
#      Also, the agent itself utilizes one node, and we consider that time as
#      agent utilization overhead.
#
# NOTE: we actually use core-seconds instead of core-hours.  So what?!
#

metrics_prte = [

        ['Agent Nodes',       ['agent']],
        ['Pilot Startup',     ['boot', 'setup_1']],
        ['Warmup',            ['warm' ]],
        ['Prepare Execution', ['exec_queue', 'exec_prep']],
        ['Execution RP',      ['exec_rp', 'exec_sh', 'term_sh', 'term_rp']],
        ['Execution PRTE',    ['prte_phase_1', 'prte_phase_2', 'prte_phase_3']],
        ['Execution Cmd',     ['exec_cmd']],
        ['Unschedule',        ['unschedule']],
        ['Draining',          ['drain']],
        ['Pilot Termination', ['term' ]],
        ['Idle',              ['idle' ]],
]


metrics_default = [

        ['Agent Nodes',       ['agent']],
        ['Pilot Startup',     ['boot', 'setup_1']],
        ['Warmup',            ['warm' ]],
        ['Prepare Execution', ['exec_queue', 'exec_prep']],
        ['Pilot Termination', ['term' ]],
        ['Execution RP',      ['exec_rp', 'exec_sh', 'term_sh', 'term_rp']],
        ['Execution Cmd',     ['exec_cmd']],
        ['Unschedule',        ['unschedule']],
        ['Draining',          ['drain']],
        ['Idle',              ['idle' ]],
]

metrics = metrics_default

metrics = [
    ['Bootstrap', ['boot', 'setup_1']                         , '#c6dbef'],
    ['Warmup'   , ['warm' ]                                   , '#f0f0f0'],
    ['Schedule' , ['exec_queue','exec_prep', 'unschedule']    , '#c994c7'],
    ['Exec RP'  , ['exec_rp', 'exec_sh', 'term_sh', 'term_rp'], '#fdbb84'],
    ['Exec Cmd' , ['exec_cmd']                                , '#e31a1c'],
    ['Cooldown' , ['drain']                                   , '#addd8e']
]

# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    if len(sys.argv) < 2:
        print("\n\tusage: %s <dir|tarball>\n" % sys.argv[0])
        sys.exit(1)

    src = sys.argv[1]

    if len(sys.argv) == 3: stype = sys.argv[2]
    else                 : stype = 'radical.pilot'

    fig, axes = plt.subplots(2, figsize=ra.get_plotsize(500))
    session   = ra.Session(src, stype=stype)

    # this script does not really work for many tasks
    tasks = session.get(etype='task')
    if len(tasks) > 1024:
        if 'RA_MANY_TASKS_OK' not in os.environ:
            print('too many tasks for this plot type.  Set "RA_MANY_TASKS_OK"')
            print('in the your environment to allow this script to run anyway.')
            sys.exit(1)

    # this script only works for one pilot
    pilots = session.get(etype='pilot')
    assert(len(pilots) == 1), len(pilots)

    sid     = session.uid
    pilot   = pilots[0]
    rm_info = pilot.cfg['resource_details']['rm_info']
    p_zero  = pilot.timestamps(event={ru.EVENT: 'bootstrap_0_start'})[0]
    p_size  = pilot.description['cores']
    n_nodes = int(p_size / rm_info['cores_per_node'])
    n_tasks = len(session.get(etype='task'))

    legend = None
    rtypes = ['cpu', 'gpu']
    for i, rtype in enumerate(rtypes):

        # get utilization information
        prov, consumed, stats_abs, stats_rel, info = session.utilization(metrics, rtype)

        with ru.ru_open('%s.stats' % sid, 'w') as fout:
            fout.write('\n%s\n\n' % info)

      # import pprint
      # pprint.pprint(prov)
      # pprint.pprint(consumed)
      # pprint.pprint(stats_abs)
      # pprint.pprint(stats_rel)

        legend, patches, x, y = ra.get_plot_utilization(metrics,
                {sid: consumed}, p_zero, sid)

        for patch in patches:
            axes[i].add_patch(patch)

        # Format axes
        axes[i].set_xlim([x['min'], x['max']])
        axes[i].set_ylim([y['min'], y['max']])

        axes[i].yaxis.set_major_locator(plt.MaxNLocator(5))
        axes[i].xaxis.set_major_locator(plt.MaxNLocator(5))

        # Resource-type dependend labels
        axes[i].set_ylabel(to_latex('%ss' % rtype.upper()))
        axes[i].set_xlabel(to_latex('time (s)'))

    # Do not repeat the X-axes label in the topmost plot
    for ax in fig.get_axes():
        ax.label_outer()

    # Title of the plot. Facultative, requires info about session
    # (see RA Info Chapter)
    axes[0].set_title(to_latex('%s Tasks - %s Nodes' % (n_tasks, n_nodes)))

    # Add legend for both plots
    fig.legend(legend, [m[0] for m in metrics], ncol=3,
               loc='upper center', bbox_to_anchor=(0.5, 1.10))


  # plt.xticks(list(range(int(x_min)-1, int(x_max)+1)))
    fig.savefig('%s_util.png' % sid, bbox_inches="tight")
  # plt.show()


# ------------------------------------------------------------------------------





#!/usr/bin/env python3

__author__    = 'RADICAL-Cybertools Team'
__email__     = 'info@radical-cybertools.org'
__copyright__ = 'Copyright 2013-23, The RADICAL-Cybertools Team'
__license__   = 'MIT'


''' Setup script, only usable via pip. '''

import os

import subprocess as sp

from glob       import glob
from setuptools import setup, Command, find_namespace_packages


# ------------------------------------------------------------------------------
#
base     = 'analytics'
name     = 'radical.%s'      % base
mod_root = 'src/radical/%s/' % base

scripts  = list(glob('bin/radical-*'))
root     = os.path.dirname(__file__) or '.'
readme   = open("%s/README.md" % root, encoding='utf-8').read()
descr    = 'Profiling library for RADICAL-Cybertools',
keywords = ['radical', 'cybertools', 'utilities', 'profiling', 'analysis']

share    = 'share/%s' % name
data     = [('%s/examples'      % share, glob('examples/*.{py,cfg,json,sh}')),
            ('%s/styles/'       % share, glob('styles/*.txt')),
            ('bin/rp_inspect'          , glob('bin/rp_inspect/*')),
]


# ------------------------------------------------------------------------------
#
def sh_callout(cmd):
    p = sp.Popen(cmd, stdout=sp.PIPE, stderr=sp.PIPE, shell=True)
    stdout, stderr = p.communicate()
    ret            = p.returncode
    return stdout, stderr, ret


# ------------------------------------------------------------------------------
#
def get_version(_mod_root):
    '''
    a VERSION file containes the version strings is created in mod_root,
    during installation.  That file is used at runtime to get the version
    information.
    '''

    _out = None
    _err = None
    _ret = None
    try:
        _version_path   = '%s/%s/VERSION' % (root, _mod_root)
        _version_base   = None
        _version_short  = None
        _version_branch = None
        _version_tag    = None
        _version_detail = None

        # get `version_base` from distribution's 'VERSION' file
        with open('%s/VERSION' % root, 'r', encoding='utf-8') as fin:
            _version_base = fin.readline().strip()

        _, _, ret = sh_callout('cd %s && git rev-parse --git-dir && which git'
                               % root)
        _in_git = (ret == 0)

        if not _in_git:

            with open(_version_path, 'w', encoding='utf-8') as fout:
                fout.write(_version_base + '\n')

        else:

            # get details from git
            _out, _err, _ret = sh_callout('cd %s && git describe --tags --always' % root)
            assert _ret == 0, 'git describe failed'
            _out = _out.decode()
            _out = _out.strip()

            _version_tag = _out

            _out, _err, _ret = sh_callout('cd %s && git branch --show-current' % root)
            assert _ret == 0, 'git branch failed'

            _out = _out.decode()
            _out = _out.strip()

            _version_branch = _out or 'detached'
            _version_branch = _version_branch.replace('detached from ', '~')

            _version_short = _version_tag.split('-')[0]
            _version_short = _version_short[1:]  # strip the 'v'

            if _version_tag:
                _version_detail = '%s-%s@%s' % (_version_base, _version_tag,
                                                _version_branch)
            else:
                _version_detail = '%s@%s' % (_version_base, _version_branch)

            with open(_version_path, 'w', encoding='utf-8') as fout:
                fout.write(_version_short  + '\n')
                fout.write(_version_base   + '\n')
                fout.write(_version_branch + '\n')
                fout.write(_version_tag    + '\n')
                fout.write(_version_detail + '\n')

        return _version_base, _version_path

    except Exception as e:
        _msg = 'Could not extract/set version: %s' % e
        if _ret:
            _msg += '\n' + _out + '\n\n' + _err
        raise RuntimeError(_msg) from e


# ------------------------------------------------------------------------------
# get version info -- this will create VERSION and srcroot/VERSION
version, version_path = get_version(mod_root)


# ------------------------------------------------------------------------------
#
class RunTwine(Command):
    user_options = []
    def initialize_options(self): pass
    def finalize_options(self):   pass
    def run(self):
        _, _, _ret = sh_callout('python3 setup.py sdist upload -r pypi')
        raise SystemExit(_ret)


# ------------------------------------------------------------------------------
#
with open('%s/requirements.txt' % root, encoding='utf-8') as freq:
    requirements = freq.readlines()


# ------------------------------------------------------------------------------
#
setup_args = {
    'name'               : name,
    'version'            : version,
    'description'        : descr,
    'long_description'   : readme,
    'long_description_content_type' : 'text/markdown',
    'author'             : 'RADICAL Group at Rutgers University',
    'author_email'       : 'radical@rutgers.edu',
    'maintainer'         : 'The RADICAL Group',
    'maintainer_email'   : 'radical@rutgers.edu',
    'url'                : 'http://radical-cybertools.github.io/%s/' % name,
    'project_urls'       : {
        'Documentation': 'https://radical%s.readthedocs.io/en/latest/' % base,
        'Source'       : 'https://github.com/radical-cybertools/%s/'   % name,
        'Issues' : 'https://github.com/radical-cybertools/%s/issues'   % name,
    },
    'license'            : 'MIT',
    'keywords'           : keywords,
    'python_requires'    : '>=3.8',
    'classifiers'        : [
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Environment :: Console',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Topic :: Utilities',
        'Topic :: System :: Distributed Computing',
        'Topic :: Scientific/Engineering',
        'Operating System :: POSIX',
        'Operating System :: Unix'
    ],
    'packages'           : find_namespace_packages('src', include=['radical.*']),
    'package_dir'        : {'': 'src'},
    'scripts'            : scripts,
    'package_data'       : {'': ['*.txt', '*.sh', '*.json', '*.gz', '*.c',
                                 '*.md', 'VERSION']},
    'install_requires'   : requirements,
    'zip_safe'           : False,
    'data_files'         : data,
    'cmdclass'           : {'upload': RunTwine},
}


# ------------------------------------------------------------------------------
#
setup(**setup_args)


# ------------------------------------------------------------------------------
# clean temporary files from source tree
os.system('rm -vrf src/%s.egg-info' % name)
os.system('rm -vf  %s'              % version_path)


# ------------------------------------------------------------------------------




#!/usr/bin/env python3

import sys
import time
import optparse                              # pylint: disable=deprecated-module

from collections import defaultdict

import radical.utils as ru


# ------------------------------------------------------------------------------
#
def usage(msg=None):

    ret = 0
    if msg:
        print('\n\n\t\tError: %s\n' % msg)
        ret = 1

    print('''

    usage  : radical-analytics-parse-profile [options] [<src>]
    example: radical-analytics-parse-profile -s 'start' -e 'start,stop' sample.prof
    source : profiler as written by `ru.Profiler`
    options:
        -h, --help                             : this message
        -n, --normalize                        : normalize timestamps
        -s, --sort-by    <event>               : sort resulting data by `event`
        -e, --events     <event_1,event_2,...> : event timestamps to extract
        -o, --output     <filename>            : output filename

    The profile will parsed for the given set of events, and the resulting data
    will be written to the given output file in the following format:


       <uid> <timestamp_1> <timestamp_2> ...

    where the timestamps correspond to the respective events.  The table will be
    sorted by the event given with `--sort-by`, if given.  If no `--sort-by` is
    given, the table will be sorted by the first timestamp column.  If
    `--normalize` is given, all timestamps will be normalized to the timestamp
    of the first event.

''')

    sys.exit(ret)


# ------------------------------------------------------------------------------
#
# parse options
#
parser = optparse.OptionParser(add_help_option=False)

parser.add_option('-h', '--help',      dest='help',      action='store_true')
parser.add_option('-n', '--normalize', dest='normalize', action='store_true')
parser.add_option('-s', '--sort-by',   dest='sort_by')
parser.add_option('-e', '--events',    dest='events')
parser.add_option('-o', '--output',    dest='output')

options, args = parser.parse_args()
if len(args) > 1:
    usage("Too many arguments (%s)" % args)

if len(args) < 1:
    src = None
else:
    src = args[0]

EVENTS = None
OUTPUT = None

if options.help  : usage()
if options.events: EVENTS = [x for x in options.events.split(',')]
if options.output: OUTPUT = options.output

data = defaultdict(dict)
tmin = None

tmin = time.time()  # assume that profiles were collected in the past

try:
    fin = None
    if src is None:
        fin = sys.stdin
    else:
        fin = ru.ru_open(src, 'r')

    for line in fin.readlines():
        elems = line.split(',')
        if EVENTS and elems[1] not in EVENTS:
            continue
        time  = float(elems[0])
        tmin  = min(tmin, time)
        uid   = elems[4]
        event = elems[1]
        data[uid][event] = time

finally:
    if fin: fin.close()


if not EVENTS:
    EVENTS = list(data[list(data.keys())[0]].keys())


if options.normalize:
    for uid in data:
        for event in data[uid]:
            data[uid][event] = data[uid][event] - tmin

try:
    if not OUTPUT:
        fout = sys.stdout
    else:
        fout = ru.ru_open(OUTPUT, 'w')

    fout.write('# %13s ' % 'uid')
    for event in EVENTS:
        fout.write('%17s ' % event)
    fout.write('\n')

    uids = list(data.keys())
    if options.sort_by:
        uids = sorted(uids, key=lambda uid: data[uid].get(options.sort_by, 0.0))

    for uid in uids:
        out = '%15s ' % uid
        for event in EVENTS:
            val = data[uid].get(event, 0.0)
            out += '%10.6f ' % val
        fout.write('%s\n' % out)

finally:
    if fout: fout.close()





from .session import Session

from typing import List, Union


# ------------------------------------------------------------------------------
#
class Experiment(object):

    # --------------------------------------------------------------------------
    #
    def __init__(self, sessions: List[Union[str, Session]],
                       stype: str = None):
        '''
        This class represents an RCT experiment, i.e., a series of RA sessions
        which are collectively analyzed.

        `sessions` is expected to be either (1) a list of session source paths
        pointing to tarballs or session directories, or (b) a list of
        `ra.Session` instances.  The order of entries in the list determines the
        default order used in plots etc.

        The session type `stype` will be uniformely applied when reading session
        data from provided paths.
        '''

        # FIXME: this is missing an abstraction: `Run`: a collection of sessions
        #         which share the same parameters and thus can be statistically
        #         handled together (means, std-deviation, etc).  Right now we
        #         only use this `Experiment` abstraction for non-statistical
        #         analysis (event plots, utilization plots, etc.)

        self._sessions = list()

        if not sessions:
            raise ValueError('cannot create experiments w/o sessions')

        if isinstance(sessions[0], str):
            for src in sessions:
                assert isinstance(src, str)
                self._sessions.append(Session.create(src=src, stype=stype))

        else:
            for session in sessions:
                assert isinstance(session, Session)
                self._sessions.append(session)


    # --------------------------------------------------------------------------
    #
    @property
    def sessions(self):
        return self._sessions

    @property
    def sids(self):
        return [s.sid for s in self._sessions]


    # --------------------------------------------------------------------------
    #
    def utilization(self, metrics, rtype='cpu', udurations=None):
        '''
        return five dictionaries:
          - provided resources
          - consumed resources
          - absolute stats
          - relative stats
          - information about resource utilization

        The resource dictionaries have the following structures::

            provided = {
                <session_id> : {
                    'metric_1' : {
                        'uid_1'        : [float, list],
                        'uid_2'        : [float, list],
                        ...
                    },
                    'metric_2' : {
                        'uid_1'        : [float, list],
                        'uid_2'        : [float, list],
                        ...
                    },
                    ...
                },
                ...
            }
            consumed = {
                <session_id> : {
                    'metric_1' : {
                        'uid_1'         : [float, list]
                        'uid_2'         : [float, list],
                        ...
                    },
                    'metric_2' :         {
                        'uid_1'         : [float, list],
                        'uid_2'         : [float, list],
                        ...
                    },
                    ...
                },
                ...
            }

        `float` is always in tasks of `resource * time`, (think `core-hours`),
        `list` is a list of 4-tuples `[t0, t1, r0, r1]` which signify at what
        specific time interval (`t0 to t1`) what specific resources (`r0 to r1`)
        have been used.  The task of the resources are here dependent on the
        session type: only RP sessions are supported at the moment where those
        resource values are indexes in to the list of cores used in that
        specific session (offset over multiple pilots, if needed).
        '''

        # FIXME: the data structure documented above is not yet implemented

        provided  = dict()
        consumed  = dict()
        stats_abs = dict()
        stats_rel = dict()
        info      = dict()

        # obtain resources provisions and consumptions for all sessions
        for session in self._sessions:

            sid = session.uid
            p, c, sa, sr, i = session.utilization(metrics, rtype, udurations)

            provided [sid] = p
            consumed [sid] = c
            stats_abs[sid] = sa
            stats_rel[sid] = sr
            info     [sid] = i

        return provided, consumed, stats_abs, stats_rel, info


    # --------------------------------------------------------------------------
    #
    def _dump_ts(self, dname, e, spec, psize=0):

        pass
      # ts   = e.timestamps(event=spec)
      # diff = ts[-1] - ts[0]
      # print '%-10s : %-55s : %3d : %10.1f - %10.1f = %10.1f -> %10.1f' \
      #     % (dname, spec, len(ts), ts[-1], ts[0], diff, diff * psize)


# ------------------------------------------------------------------------------




#!/usr/bin/env python

__copyright__ = 'Copyright 2013-2016, http://radical.rutgers.edu'
__license__   = 'MIT'


import sys

import radical.utils     as ru
import radical.analytics as ra

import matplotlib.pyplot as plt
import numpy             as np


# We plot timelines for all events listed in `event_list` for all entities of
# type `event_entity`..  Before plotting, we sort those entities by the
# timestamp of the first event in the event list

event_entity = 'unit'
event_list   = \
    [
      # {ru.STATE: 'NEW'                         , ru.EVENT: 'state'          },
      # {ru.STATE: 'UMGR_SCHEDULING_PENDING'     , ru.EVENT: 'state'          },
      # {ru.STATE: 'UMGR_SCHEDULING'             , ru.EVENT: 'state'          },
      # {ru.STATE: 'UMGR_STAGING_INPUT_PENDING'  , ru.EVENT: 'state'          },
      # {ru.STATE: 'UMGR_STAGING_INPUT'          , ru.EVENT: 'state'          },
      # {ru.STATE: 'AGENT_STAGING_INPUT_PENDING' , ru.EVENT: 'state'          },
      # {ru.COMP : 'agent_0'                     , ru.EVENT: 'get'            },
        {ru.STATE: 'AGENT_STAGING_INPUT'         , ru.EVENT: 'state'          },
        {ru.STATE: 'AGENT_SCHEDULING_PENDING'    , ru.EVENT: 'state'          },
        {ru.STATE: 'AGENT_SCHEDULING'            , ru.EVENT: 'state'          },
      # {ru.STATE: None                          , ru.EVENT: 'schedule_ok'    },
        {ru.STATE: 'AGENT_EXECUTING_PENDING'     , ru.EVENT: 'state'          },
        {ru.STATE: 'AGENT_EXECUTING'             , ru.EVENT: 'state'          },
      # {ru.STATE: None                          , ru.EVENT: 'exec_mkdir'     },
      # {ru.STATE: None                          , ru.EVENT: 'exec_mkdir_done'},
      # {ru.STATE: None                          , ru.EVENT: 'exec_start'     },
      # {ru.STATE: None                          , ru.EVENT: 'app_start'      },
      # {ru.STATE: None                          , ru.EVENT: 'app_stop'       },
      # {ru.STATE: None                          , ru.EVENT: 'exec_ok'        },
      # {ru.STATE: None                          , ru.EVENT: 'exec_stop'      },
        {ru.STATE: 'AGENT_STAGING_OUTPUT_PENDING', ru.EVENT: 'state'          },
        {ru.STATE: 'AGENT_STAGING_OUTPUT'        , ru.EVENT: 'state'          },
      # {ru.STATE: 'UMGR_STAGING_OUTPUT_PENDING' , ru.EVENT: 'state'          },
      # {ru.STATE: 'UMGR_STAGING_OUTPUT'         , ru.EVENT: 'state'          },
      # {ru.STATE: 'DONE'                        , ru.EVENT: 'state'          },
    ]


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    if len(sys.argv) < 2:
        print("\n\tusage: %s <dir|tarball>\n" % sys.argv[0])
        sys.exit(1)

    src = sys.argv[1]

    if len(sys.argv) == 2: stype = 'radical.pilot'
    else                 : stype = sys.argv[2]

    session = ra.Session.create(src, stype)

    # A formatting helper before starting...
    def ppheader(message):
        separator = '\n' + 78 * '-' + '\n'
        print(separator + message + separator)

    data = dict()
    pipe = dict()

    for thing in session.get(etype=event_entity):

        tstamps = list()

        for event in event_list:
            times = thing.timestamps(event=event)
            if times: tstamps.append(times[0])
            else    : tstamps.append(None)

        data[thing.uid] = tstamps
        pipe[thing.uid] = [int(x) for x
                           in thing.description.get('name', '').split()]

  # diffs = list()
  # for uid in data:
  #     diffs.append(data[uid][-1] - data[uid][0])
  # print sorted(diffs)


  # sort x-axis (unit IDs) by
  #     'uid'  : task ID
  #     'get'  : time of ingest
  #     'sched': time of sccheduling
  #     'pipe' : pipeline ID (RE sessions only)
    order = 'AGENT_EXECUTING state'
    index = 4

    if order == 'uid':
        sorted_uids = sorted(pipe.keys())
    else:
        sorted_uids = [x[0] for x in sorted(list(data.items()),
                                            key=lambda v: v[1][index])]
    sorted_data   = list()
    index         = 0
    for uid in sorted_uids:
        sorted_data.append([index] + data[uid])
        index += 1

    np_data = np.array(sorted_data)

    plt.figure(figsize=(10,7))
    for e_idx in range(len(event_list)):
        plt.plot(np_data[:,0], np_data[:,(1 + e_idx)],
                 label=ru.event_to_label(event_list[e_idx]))
    plt.xlabel('task ID')
    plt.ylabel('time [sec]')

    plt.xlabel('task (sorted by %s)' % order)
    plt.ylabel('time / sec')
    plt.legend(loc='upper center', bbox_to_anchor=(0.5, 1.15),
               ncol=2, fancybox=True, shadow=True)
    plt.savefig('%s.state.png' % session.uid)
    plt.show()


# ------------------------------------------------------------------------------




#!/usr/bin/env python

import sys
import pprint

import radical.analytics as ra

__copyright__ = 'Copyright 2013-2016, http://radical.rutgers.edu'
__license__   = 'MIT'

"""
This example illustrates the use of the method ra.Session.filter()
"""

# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    if len(sys.argv) < 2:
        print("\n\tusage: %s <dir|tarball>\n" % sys.argv[0])
        sys.exit(1)

    src = sys.argv[1]

    if len(sys.argv) == 2: stype = 'radical.pilot'
    else                 : stype = sys.argv[2]

    session = ra.Session.create(src, stype)

    # A formatting helper before starting...
    def ppheader(message):
        separator = '\n' + 78 * '-' + '\n'
        print(separator + message + separator)

    # we want to focus on pilots and units
    ppheader("Filter 'unit' and 'pilot' entities")
    session.filter(etype=['unit', 'pilot'], inplace=True)
    pprint.pprint(session.list(pname='uid'))

    # for all pilots, we want to:
    #   - print the resource they have been running on
    #   - print the set of unit UIDs they have been executing
    ppheader("show pilot-to-unit mapping")
    pprint.pprint(session.describe('relations', ['pilot', 'unit']))

    ppheader("show pilot-to-resource mapping")
    for pilot in session.get(etype=['pilot']):
        print('%s : %-35s : %s' % (pilot.uid,
                                   pilot.description['resource'],
                                   pilot.cfg['hostid']))





import os
import sys
import glob

import pandas        as pd
import numpy         as np
import matplotlib    as mpl

import radical.utils as ru


# ------------------------------------------------------------------------------
#
def get_plotsize(width, height=None, fraction=1, subplots=(1, 1)):
    """ Sets aesthetic figure dimensions to avoid scaling in latex.

    Parameters
    ----------
    width   : float
              Width in points (pts).
    fraction: float
              Fraction of the width and hight the figure to occupies
    subplots: tuple
              Number of raws and number of columns of the plot.

    Returns
    -------
    fig_dim : tuple
              Dimensions of figure in inches.
    """
    # Width of figure
    fig_width_pt = width * fraction

    # Convert from pt to inches
    inches_per_pt = 1 / 72.27

    # Figure width in inches
    fig_width_in = fig_width_pt * inches_per_pt

    # Figure height in inches
    if height:
        fig_height_in = height * fraction * inches_per_pt

    else:
        # Golden ratio to set aesthetic figure height
        golden_ratio  = (5 ** 0.5 - 1) / 2
        fig_height_in = fig_width_in * golden_ratio * (subplots[0] / subplots[1])

    return fig_width_in, fig_height_in


# ------------------------------------------------------------------------------
#
def get_mplstyle(name):
    """Returns the installation path of a Matplotlib style.

    Parameters
    ----------
    name: string
          Filename ending in .txt.

    Returns
    -------
    path : string
           Normalized path.
    """

    path  = os.path.dirname(sys.executable)
    path += '/../share/radical.analytics/styles'

    for path in glob.glob('%s/*.txt' % path):
        if path.endswith('/%s.txt' % name):
            return os.path.normpath(path)


    raise RuntimeError('style %s not found' % name)


# ------------------------------------------------------------------------------
#
def stack_transitions(series, tresource, to_stack):
    '''Creates time series for each metric and combines them into one data frame
    for alignment. Since transitions obviously happen at arbitrary times, the
    timestamps for metric A may see no transitions for metric B. When using a
    combined timeline, we end up with NaN entries for some metrics on most
    timestamp, which in turn leads to gaps when plotting. So we fill the NaN
    values with the previous valid value, which in our case holds until the next
    transition happens.

    Parameters
    ----------
    series   : dict
               Pairs of timestamps for each metric of each type of
               resource. E.g. series['cpu']['term'] = [[0.0, 0.0],
               [302.4374113082886, 100.0], [304.6761999130249, 0.0]].
    tresource: string
               Type of resource. E.g., 'cpu' or 'gpu'.

    to_stack : list
               List of metrics to stack. E.g., ['bootstrap', 'exec_cmd',
               'schedule', 'exec_rp', 'term', 'idle'].

    Returns
    -------
    stacked : pandas.DataFrame
              Columns: time and one for each metric. Rows: timestamp and
              percentage / amount of resource utilization for each metric at
              that point in time.
    '''

    # find the global time line
    glob_times = set()
    for m,df in series[tresource].items():
        for t,_ in df:
            glob_times.add(t)

    glob_times = sorted(glob_times)

    # create a timeline for each metric, set missing values as NaN
    tlines = dict()
    tlines['time'] = glob_times
    for m in series[tresource]:
        tlines[m] = [np.nan] * len(glob_times)

        for t,v in series[tresource][m]:
            t_idx = glob_times.index(t)
            tlines[m][t_idx] = v

    # create dataframe and fill all NaN values with the previous valid value
    merged = pd.DataFrame(tlines)
    merged.ffill(inplace=True)

    # stacked plotting and area filling don't play well together in matplotlib.
    # Instead we use normal (unstacked) plot routines and fill in between, we
    # manually compute the stacked numbers: copy the timeline to a new data
    # frame.
    stacked = merged[['time']].copy()
    prev    = list()

    # for each metric, copy the metric column and add all previous columns
    for m in to_stack:
        stacked[m] = merged[m]
        for p in prev:
            stacked[m] += merged[p]
        prev.append(m)

    return stacked


# ------------------------------------------------------------------------------
#
def get_pilot_series(session, pilot, tmap, resrc, percent=True):
    """ Derives the series of pilot resource transition points from the metrics.

    Parameters
    ----------
    session: ra.Session
             The Session object of RADICAL-Analytics created from a RCT sandbox.
    pilot  : ra.Entity
             The pilot object of session.
    tmap   : dict
             Map events to transition points in which a metric changes its
             owner. E.g., [{1: 'bootstrap_0_start'}, 'system', 'bootstrap']
             defines bootstrap_0_start as the event in which resources pass
             from the system to the bootstrapper.
    resrc  : list
             Type of resources. E.g., ['cpu', 'gpu'].
    percent: bool
             Whether we want to return resource utilization as percentage of
             the total resources available or as count of a type of resource.

    Returns
    -------
    p_resrc: dict
             Amount of resources in the pilot.
    series : dict
             List of time series per metric and resource type. E.g.,
             series['cpu']['term'] = [[0.0, 0.0], [302.4374113082886, 100.0],
             [304.6761999130249, 0.0]].
    x      : dict
             Mix and max value of the X-axes.

    """

    # get total pilot resources and runtime,
    p_resrc = {'cpu': pilot.cfg['cores'],
               'gpu': pilot.cfg['gpus' ]}

    t_min = None
    t_max = None

    try   : t_min = pilot.timestamps(event={1: 'bootstrap_0_start'})[0]
    except: pass

    try   : t_max = pilot.timestamps(event={1: 'bootstrap_0_stop'})[0]
    except: pass

    # fallback for missing bootstrap events
    if t_min is None: t_min = pilot.timestamps(state='PMGR_ACTIVE')[0]
    if t_max is None: t_max = pilot.events[-1][ru.TIME]

    assert t_min is not None
    assert t_max is not None

    t_span = t_max - t_min
    x_min  = 0
    x_max  = t_span + 0.05 * t_span

    # derive the pilot resource transition points from the metrics. Metrics
    # might be pulled from rp.utils but should be consistent with those used for
    # RU v.1.
    # rpp = rp.utils.prof_utils

    # get all contributions
    metrics = list()
    for trans in tmap.values():
        metrics += [x[1] for x in trans]
    metrics  = set(metrics)

    # prepare the contributions data structure which gets filled below
    contribs = {r: {
                m: [[0.0, 0.0]]
                   for m in metrics}
                for r in resrc}

    for entity in session.get():

        td    = entity.description
        etype = entity.etype

        transitions = tmap.get(etype)
        if not transitions:
          # print('no transitions for %s: etype %s' % (entity.uid, etype))
            continue

      # print('\n', entity.uid, etype, sorted(set([e[1] for e in entity.events])))
      # print()

        for trans in transitions:

            event  = trans[0]
            p_from = trans[1]
            p_to   = trans[2]

            try:
                t_resrc = {'cpu': entity.resources.get('cpu'),
                           'gpu': entity.resources.get('gpu')}

                # raptor tasks which were created in the master may not have
                # resources defined, or may not even have a task description
                # (they don't get an entry on MongoDB).  We thus guess their
                # resource consumption here.
                # NOTE: this can lead to underutilization to be reported!
                if not t_resrc['cpu'] and 'task' in entity.uid:
                    td = entity.description
                    if not td.get('cpu_processes'):
                        # guess
                        t_resrc = {'cpu': 1, 'gpu': 0}
                    else:
                        cores = td['cpu_processes'] * td['cpu_threads']
                        gpus  = td['cpu_processes'] * td['gpu_processes']
                        t_resrc = {'cpu': cores,
                                   'gpu': gpus}
              # print(entity.uid, p_from, p_to, t_resrc)

            except Exception as e:
                # if 'request' not in entity.uid:
                #     print('guess resources for %s' % entity.uid)

                # if 'pilot' in entity.uid:
                #     t_resrc = {'cpu': 1024 * 40,
                #                'gpu': 1024 *  8}
                # else:
                #     t_resrc = {'cpu': 1,
                #                'gpu': 0}
                raise RuntimeError(entity.uid + 'is missing resource '
                    'information. RA cannot know how many cores/GPUs were '
                    'requested. Session is likely corrupted. Aborting.') from e

            # we need to work around the fact that sub-agents have no separate
            # entity type, but belong to the pilot.  So instead we assign them
            # resources of 1 node.  We take those data from the pilot.
            if 'agent' in str(event):
                t_resrc = {'cpu': pilot.cfg['cores_per_node'],
                           'gpu': pilot.cfg['gpus_per_node' ]}

            ts = entity.timestamps(event=event)
            if not ts:
              # print('%s: no event %s for %s' % (uid, event, etype))
                continue

            for r in resrc:
                amount = t_resrc[r]
                if amount == 0:
                    continue
                t = (ts[0] - t_min)
                contribs[r][p_from].append([t, -amount])
                contribs[r][p_to  ].append([t, +amount])
              # print('%6.3f : %-30s : %-25s : %-15s --> %-15s [%s]' %
              #         (t, uid, event, p_from, p_to,   amount))

    # we now have, for all metrics, a list of resource changes, in the form of
    #
    #   [timestamp, change]
    #
    # where the change can be positive or negative.  From this, we now calculate
    # the continuous time series for the metrics: for each metric, sort the
    # contribution changes by time and calculate the running sum of the changes.

    series = dict()
    for r in contribs:

        series[r] = dict()
        for m in contribs[r]:

            series[r][m] = list()
            value = 0.0

            for c in sorted(contribs[r][m]):
                value += c[1]
                # normalize to pilot resources to obtain percent
                if p_resrc[r]:
                    if percent:
                        rel = value / p_resrc[r] * 100
                      # print('rel', rel)
                        series[r][m].append([c[0], rel])
                    else:
                        series[r][m].append([c[0], value])
                      # print('val %6.3f %-15s: %3d' % (c[0], m, value))
                else:
                    series[r][m].append([c[0], 0])

    x = {'min': x_min, 'max': x_max}

  # import pprint
  # pprint.pprint(p_resrc)
  # pprint.pprint(series)
  # pprint.pprint(x)

    return p_resrc, series, x


# ------------------------------------------------------------------------------
#
def get_pilots_zeros(ra_exp_obj):
    """Calculates when a set of pilots become available.

    Parameters
    ----------
    ra_exp_obj: ra.Experiment
                RADICAL-Analytics Experiment object with all the pilot entity
                objects for which to calculate the starting timestamp.

    Returns
    -------
    p_zeros: dict
             Session ID, pilot ID and starting timestamp. E.g.,
             {'re.session.login1.lei.018775.0005': {'pilot.0000':
             2347.582849740982}}.
    """

    p_zeros = {}
    for session in ra_exp_obj.sessions:
        p_zeros[session.uid] = {}
        for pilot in session.get(etype='pilot'):
            p_zeros[session.uid][pilot.uid] = pilot.timestamps(
                event={ru.EVENT: 'bootstrap_0_start'})[0]

    return p_zeros


# ------------------------------------------------------------------------------
#
def get_plot_utilization(metrics, consumed, t_zero, sid):
    """Calculates the resources utilized by a set of metrics. Utilization is
    calculated for each resource without stacking and aggregation. May take
    hours or days with >100K tasks, 100K resource items. Use get_pilot_series
    and stack_transitions instead.

    Parameters
    ----------
    metrics : list
              Each element is a list with name, metrics and color. E.g.,
              ['Bootstrap', ['boot', 'setup_1'], '#c6dbef'].
    consumed: dict
              min-max timestamp and resource id range for each metric and
              pilot. E.g., {'boot': {'pilot.0000': [[2347.582849740982,
              2365.6164498329163, 0, 167]}.
    t_zero  : float
              Start timestamp for the pilot.
    sid     : string
              Identifier of a ra.Session object.

    Returns
    -------
    legend : dict
             keys: Type of resource ('cpu', 'gpu'); values: list of
             matplotlib.lines.Line2D objects for the plot's legend.
    patches: dict
             keys: Type of resource ('cpu', 'gpu'); values: list of
             matplotlib.patches.Rectangle. Each rectangle represents the
             utilization for a set of resources.
    x      : dict
             Mix and max value of the X-axes.
    y      : dict
             Mix and max value of the Y-axes.
    """

    legend = list()

    x_min = None
    x_max = None
    y_min = None
    y_max = None

    patches = []

    for metric in metrics:

        color = metric[2]

        legend.append(mpl.lines.Line2D([0], [0], color=color, lw=6))

        if isinstance(metric, list):
            parts = metric[1]
        else:
            parts = [metric]

        for part in parts:
            for uid in consumed[sid][part]:
                for block in consumed[sid][part][uid]:
                    orig_x = block[0] - t_zero
                    orig_y = block[2] - 0.5
                    width  = block[1] - block[0]
                    height = block[3] - block[2] + 1.0

                    if x_min is None:
                        x_min = orig_x
                    if x_max is None:
                        x_max = orig_x + width
                    if y_min is None:
                        y_min = orig_x
                    if y_max is None:
                        y_max = orig_x + height

                    x_min = min(x_min, orig_x)
                    y_min = min(y_min, orig_y)
                    x_max = max(x_max, orig_x + width)
                    y_max = max(y_max, orig_y + height)

                    patch = mpl.patches.Rectangle((orig_x, orig_y),
                                                  width, height,
                                                  facecolor=color,
                                                  edgecolor='black',
                                                  fill=True, lw=0.0)

                    patches.append(patch)

    x = {'min': x_min, 'max': x_max}
    y = {'min': y_min, 'max': y_max}

    return legend, patches, x, y


# ------------------------------------------------------------------------------
#
def to_latex(data):
    '''
    Transforms the input string(s) so that it can be used as latex compiled plot
    label, title etc. Escapes special characters with a slash.

    Parameters
    ----------
    data : list or str
           An individual string or a list of strings to transform.

    Returns
    -------
    data : list of str
           Transformed data.
    '''

    if isinstance(data, list):
        return [to_latex(x) for x in data]

    if isinstance(data, str):
        return data.replace('%',  '\\%') \
                   .replace('#',  '\\#') \
                   .replace('_',  '\\_') \
                   .replace('$',  '\\$') \
                   .replace('&',  '\\&') \
                   .replace('~',  '\\~') \
                   .replace('^',  '\\^') \
                   .replace('{',  '\\{') \
                   .replace('}',  '\\}')

    return to_latex(str(data))


# ------------------------------------------------------------------------------
#
def tabulate_durations(durations):
    '''
    Takes a dict of durations as defined in rp.utils (e.g.,
    `rp.utils.PILOT_DURATIONS_DEBUG`) and returns a list of durations with their
    start and stop timestamps. That list can be directly converted to a
    panda.df.

    Parameters
    ----------
    durations : dict
                Dict of lists of dicts/lists of dicts. It contains
                details about states and events.

    Returns
    -------
    data : list
           List of dicts, each dict containing 'Duration Name',
           'Start Timestamp' and 'Stop Timestamp'.
    '''
    table = []
    for name in durations:
        duration = {}
        start = durations[name][0]
        stop  = durations[name][1]

        duration['Duration Name'] = name

        if list(start.values())[0] == 'state':
            duration['Start Timestamp'] = list(start.values())[1]
        else:
            duration['Start Timestamp'] = list(start.values())[0]

        if isinstance(stop, list):
            ds = []
            for state in stop:
                ds.append(list(state.values())[1])
            duration['Stop Timestamp'] = ', '.join(map(str, ds))
        else:
            if list(stop.values())[0] == 'state':
                duration['Stop Timestamp'] = list(stop.values())[1]
            else:
                duration['Stop Timestamp'] = list(stop.values())[0]

        table.append(duration)

    return table




import sys
import pprint

import radical.utils as ru


# ------------------------------------------------------------------------------
#
class Entity(object):

    def __init__(self, _uid, _profile, _details):
        """
        Args:
            uid (`str`): an ID assumed to be unique in the scope of an RA
                Session
            profile: a list of profile events for this entity
            details: a dictionary of complementary information on this entity
        """

        assert _uid
        assert _profile

        self._uid         = _uid
        self._details     = _details
        self._etype       = self._details.get('etype')
        self._description = self._details.get('description', dict())
        self._cfg         = self._details.get('cfg',         dict())
        self._resources   = self._details.get('resources',   dict())

        self._name        =  self._description.get('name') \
                          or self._description.get('job_name') \
                          or self._uid

        # if have no etype tree information, guess the etype from uid
        if not self._etype:
            self._etype = self._uid.split('.')[0]

        # FIXME: this should be sorted out on RP level
        self._cfg['hostid'] = self._details.get('hostid')

        # entities for which we have no tree information are raptor tasks (they
        # were created by the master and never saw the client side)
        # FIXME: this should be sorted out on RP level
        if not self._etype:
            if not self._details and 'task' in self._uid:
                self._etype = 'raptor.task'
            else:
                self._etype = 'unknown'

        self._states      = dict()
        self._events      = list()
        self._consistency = {'log'         : list(),
                             'state_model' : None,
                             'event_model' : None,
                             'timestamps'  : None}

        self._t_start     = None
        self._t_stop      = None
        self._ttc         = None

        self._initialize(_profile)


    # --------------------------------------------------------------------------
    #
    def __getstate__(self):

        state = {
                 'uid'         : self._uid,
                 'name'        : self._name,
                 'etype'       : self._etype,
                 'details'     : self._details,
                 'description' : self._description,
                 'resources'   : self._resources,
                 'cfg'         : self._cfg,

                 'states'      : self._states,
                 'events'      : self._events,
                 'consistency' : self._consistency,

                 't_start'     : self._t_start,
                 't_stop'      : self._t_stop,
                 'ttc'         : self._ttc,
                }

        return state


    # --------------------------------------------------------------------------
    #
    def __setstate__(self, state):

        self._uid          = state['uid']
        self._etype        = state['etype']
        self._details      = state['details']
        self._description  = state['description']
        self._resources    = state['resources']
        self._cfg          = state['cfg']

        self._states       = state['states']
        self._events       = state['events']
        self._consistency  = state['consistency']

        self._t_start      = state['t_start']
        self._t_stop       = state['t_stop']
        self._ttc          = state['ttc']

        self._name         =  self._description.get('name') \
                           or self._description.get('job_name') \
                           or self._uid


    # --------------------------------------------------------------------------
    #
    @property
    def t_start(self):
        return self._t_start

    @property
    def t_stop(self):
        return self._t_stop

    @property
    def ttc(self):
        return self._ttc

    @property
    def t_range(self):
        return [self._t_start, self._t_stop]

    @property
    def uid(self):
        return self._uid

    @property
    def name(self):
        return self._name

    @property
    def etype(self):
        return self._etype

    @property
    def states(self):
        return self._states

    @property
    def description(self):
        return self._description

    @property
    def resources(self):
        return self._resources

    @property
    def cfg(self):
        return self._cfg

    @property
    def events(self):
        return self._events

    @property
    def consistency(self):
        return self._consistency


    # --------------------------------------------------------------------------
    #
    def __str__(self):

        return "ra.Entity [%s]: %s\n    states: %s" \
                % (self.etype, self.uid, list(self._states.keys()))


    # --------------------------------------------------------------------------
    #
    def __repr__(self):

        return str(self)


    # --------------------------------------------------------------------------
    #
    def _initialize(self, profile):

        # only call once
        assert not self._states
        assert not self._events

        if profile:
            self._t_start = sys.float_info.max
            self._t_stop  = sys.float_info.min

        # we expect each event tuple to have `time` and `event`, and expect
        # 'advance' events to signify a state transition.
        for event in sorted(profile, key=lambda x: (x[ru.TIME])):

            t = event[ru.TIME]

            self._t_start = min(self._t_start, t)
            self._t_stop  = max(self._t_stop,  t)

            # FIXME: this should be phased out
            if event[ru.EVENT] in 'advance':
                event[ru.EVENT] = 'state'

            if event[ru.EVENT] == 'state':
                state = event[ru.STATE]
                self._states[state] = event

            # we also treat state transitions as generic event.
            # Because, why not?
            self._events.append(event)

        if profile:
            self._ttc = self._t_stop - self._t_start

        # FIXME: assert state model adherence here (if state model is defined)


    # --------------------------------------------------------------------------
    #
    def _ensure_tuplelist(self, events):

        if not events:
            return []

        ret = list()
        if not isinstance(events, list):
            events = [events]

        for e in events:
            if isinstance(e,dict):
                et = ru.PROF_KEY_MAX * [None]
                for k,v in list(e.items()):
                    et[k] = v
                ret.append(tuple(et))
            else:
                ret.append(e)

        return ret


    # --------------------------------------------------------------------------
    #
    def as_dict(self):

        return {
                'uid'        : self._uid,
                'etype'      : self._etype,
                'states'     : self._states,
                'events'     : self._events,
                'cfg'        : self._cfg,
                'resources'  : self._resources,
                'description': self._description,
               }


    # --------------------------------------------------------------------------
    #
    def dump(self):

        pprint.pprint(self.as_dict())


    # --------------------------------------------------------------------------
    #
    def list_states(self):

        return list(self._states.keys())


    # --------------------------------------------------------------------------
    #
    def duration(self, state=None, event=None, time=None, ranges=None):
        """
        This method accepts a set of initial and final conditions, interprets
        them as documented in the `ranges()` method (which has the same
        signature), and then returns the difference between the resulting
        timestamps.
        """

        if not ranges:
            ranges = self.ranges(state, event, time)
          # print 'get %5d ranges for %s' % (len(ranges), self.uid)
          # pprint.pprint(self.events)

        else:
            assert not state
            assert not event
            assert not time

            # make sure the ranges are collapsed (although they likely are
            # already...)
          # print 'use %5d ranges for %s' % (len(ranges), self.uid)
            ranges = ru.collapse_ranges(ranges)

        if not ranges:
            raise ValueError('no duration defined for given constraints '
                  '(%s) (%s) (%s) (%s)' % (state, event, time, ranges))

        return sum(r[1] - r[0] for r in ranges)


    # --------------------------------------------------------------------------
    #
    def timestamps(self, state=None, event=None, time=None):
        """
        This method accepts a set of conditions, and returns the list of
        timestamps for which those conditions applied, i.e., for which state
        transitions or events are known which match the given 'state' or 'event'
        parameter.  If no match is found, an empty list is returned.

        Both `state` and `event` can be lists, in which case the union of all
        timestamps are returned.

        The `time` parameter is expected to be a single tuple, or a list of
        tuples, each defining a pair of start and end time which are used to
        constrain the matching timestamps.

        The returned list will be sorted.
        """

        event = self._ensure_tuplelist(event)
        state = ru.as_list(state)
        ret   = list()

        if not event and not state:
            # no filters, consider all events
            ret = self._events

        for e in event:
            for x in self._events:
                if self._match_event(e,x):
                    ret.append(x[ru.TIME])

        for s in state:
            if s in self._states:
                ret.append(self._states[s][ru.TIME])

        # apply time filters
        if time:
            matched = list()
            for etime in ret:
                for ttuple in time:
                    if etime >= ttuple[0] and etime <= etime[1]:
                        matched.append(etime)
                        break
            ret = matched

        return sorted(ret)


    # --------------------------------------------------------------------------
    #
    def _match_event(self, needle, hay):

        for key in range(ru.PROF_KEY_MAX - 1):
            if needle[key] is not None:
                if key == ru.MSG:
                    if needle[key] not in hay[key]:
                        return False
                else:
                    if needle[key] != hay[key]:
                        return False
        return True


    # --------------------------------------------------------------------------
    #
    def ranges(self, state=None, event=None, time=None,
                     expand=False, collapse=True):
        """
        This method accepts a set of initial and final conditions, in the form
        of range of state and or event specifiers::

          entity.ranges(state=[['INITIAL_STATE_1', 'INITIAL_STATE_2'],
                                'FINAL_STATE_1',   'FINAL_STATE_2'  ]],
                        event=[[ initial_event_1,   initial_event_2 ]
                               [ final_event_1,     final_event_2   ]],
                        time =[[2.0, 2.5], [3.0, 3.5]])

        More specifically, the `state` and `event` parameter are expected to be
        a tuple, where the first element defines the initial condition, and the
        second element defines the final condition.  The `time` parameter is
        expected to be a single tuple, or a list of tuples, each defining a pair
        of start and end time which are used to constrain the resulting ranges.
        States are expected as strings, events as full event tuples::

            [ru.TIME,  ru.NAME, ru.UID,  ru.STATE, ru.EVENT, ru.MSG,  ru.ENTITY]

        where empty fields are not applied in the filtering - all other fields
        must match exactly.  The events can also be specified as dictionaries,
        which then don't need to have all fields set.

        The method will:

          - determine the *earliest* timestamp when any of the given initial
            conditions have been met, which can be either an event or a state;
          - determine the *next* timestamp when any of the given final
            conditions have been met (when `expand` is set to `False` [default])
            OR
          - determine the *last* timestamp when any of the given final
            conditions have been met (when `expand` is set to `True`)

        From that final point in time the search for the next initial condition
        applies again, which may result in another time range to be found.  The
        method returns the set of found ranges, as a list of `[start, end]` time
        tuples.

        The resulting ranges are constrained by the `time` constraints, if such
        are given.

        Note that with `expand=True`, at most one range will be found.

        Setting 'collapse' to 'True' (default) will prompt the method to
        collapse the resulting set of ranges.

        The returned ranges are time-sorted

        Example::

           task.ranges(state=[rp.NEW, rp.FINAL]))
           task.ranges(event=[{ru.NAME : 'exec_start'},
                              {ru.NAME : 'exec_ok'}])
        """

        # NOTE: this method relies on all state changes (as events in
        #       `self.states`) to also be recorded as events (as events in in
        #       `self.events` with `ru.NAME == 'state'`).

        if not state and not event:
            raise ValueError('duration needs state and/or event arguments')

        event = self._ensure_tuplelist(event)

        if not state: state = [[], []]
        if not event: event = [[], []]

        s_init  = state[0]
        s_final = state[1]
        e_init  = event[0]
        e_final = event[1]

        if not isinstance(s_init,  list): s_init  = [s_init ]
        if not isinstance(s_final, list): s_final = [s_final]
        if not isinstance(e_init,  list): e_init  = [e_init ]
        if not isinstance(e_final, list): e_final = [e_final]

        conds_init  = list()
        conds_final = list()

        for s in s_init:
            et = ru.PROF_KEY_MAX * [None]
            et[ru.STATE] = s
            et[ru.EVENT] = 'state'
            conds_init.append(tuple(et))

        for s in s_final:
            et = ru.PROF_KEY_MAX * [None]
            et[ru.STATE] = s
            et[ru.EVENT] = 'state'
            conds_final.append(tuple(et))

        for e in e_init:
            if isinstance(e,dict):
                et = ru.PROF_KEY_MAX * [None]
                for k,v in list(e.items()):
                    et[k] = v
                conds_init.append(tuple(et))
            else:
                conds_init.append(e)

      # t_start = sys.float_info.max
      # for e in e_init:
      #     for e_info in self._events:
      #         if self._match_event(e, e_info):
      #             t_start = min(t_start, e_info[ru.TIME])
      #
      # t_stop  = sys.float_info.min
      # for e in e_final:
      #     for e_info in self._events:
      #         if self._match_event(e, e_info):
      #             t_stop = max(t_stop, e_info[ru.TIME])
      #
      # if t_start == sys.float_info.max:
      #   # return []
      #     raise ValueError('initial condition did not apply')
      #
      # if t_stop == sys.float_info.min:
      #   # return []
      #     raise ValueError('final condition did not apply')
      #
      # if t_stop < t_start:
      #   # return []
      #     raise ValueError('duration uncovered time inconsistency')


        for e in e_final:
            if isinstance(e,dict):
                et = ru.PROF_KEY_MAX * [None]
                for k,v in list(e.items()):
                    et[k] = v
                conds_final.append(tuple(et))
            else:
                conds_final.append(e)

        ranges     = list()
        this_range = [None, None]

        # NOTE: this assumes that `self.events` are time sorted
        for e in self._events:
            if this_range[0] is None:
                # check for an initial event.
                for c in conds_init:
                    if self._match_event(c, e):
                        this_range[0] = e[ru.TIME]
                        break
            else:
                # check for a final event.  If found and '!expand`, then store
                # the now completed event away, and start a new one; if
                # `expand`, then keep searching for a later final event
                for c in conds_final:
                    if self._match_event(c, e):
                        this_range[1] = e[ru.TIME]
                        if not expand:
                            ranges.append(this_range)
                            this_range = [None, None]
                            break

        # we went through all events.  `this_range` may or may not be a usable
        # range here.  If it is, append it to ranges.
        if  this_range[0] is not None and \
            this_range[1] is not None     :
            ranges.append(this_range)

        # apply time filter, if specified
        # For all ranges, check if they fall completely or partially within any
        # of the given time filters.  If not, drop that range, if yes, include
        # the overlapping part.
        if not time or not len(time):
            ret = ranges

        else:
            ret = list()
            if not isinstance(time[0], list):
                time = [time]

            # for each range in ret, we make  sure that it does not violate any
            # time filter
            for erange in ranges:
                for trange in time:
                    new_start = max(trange[0], erange[0])
                    new_stop  = min(trange[1], erange[1])
                    if new_stop > new_start:
                        ret.append([new_start, new_stop])

        if collapse:
            ret = ru.collapse_ranges(ret)

        return sorted(ret)


# ------------------------------------------------------------------------------




#!/usr/bin/env python

import sys
import pprint

import radical.pilot as rp
import radical.analytics as ra

__copyright__ = 'Copyright 2013-2016, http://radical.rutgers.edu'
__license__   = 'MIT'

"""
This example illustrates the use of the method ra.Session.filter()
"""

# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    if len(sys.argv) < 2:
        print("\n\tusage: %s <dir|tarball>\n" % sys.argv[0])
        sys.exit(1)

    src = sys.argv[1]

    if len(sys.argv) == 2: stype = 'radical.pilot'
    else                 : stype = sys.argv[2]

    session = ra.Session.create(src, stype)

    # A formatting helper before starting...
    def ppheader(message):
        separator = '\n' + 78 * '-' + '\n'
        print(separator + message + separator)

    # and here we go. As seen in example 02, we use ra.Session.get() to get all
    # the entity objects with one or more type, uid, or state. Runs with
    # thousands of entities produce an amount of data large enough to slow down
    # the analysis, once loaded into the ra.session object. ra.Session.filter()
    # enables to reduce the size of the session object by keeping only the data
    # that are relevant to our analysis.
    #
    # For example, when using RADICAL-Pilot, we can keep only the data relative
    # to entities of type 'unit' and 'pilot':
    ppheader("Filter 'unit' and 'pilot' entities")
    units_and_pilots = session.filter(etype=['unit', 'pilot'], inplace=False)
    pprint.pprint(units_and_pilots.get())

    # Still quite a lot of data. If our analysis is exploratory, we may be
    # interested only to know how many entities have failed:
    ppheader("Filter 'unit' and 'pilot' entities with a rp.FAILED state")
    units_pilots_start_end = session.filter(etype=['unit', 'pilot'],
                                            state=[rp.DONE],
                                            inplace=False)
    pprint.pprint(units_and_pilots.list(['etype', 'state']))

    # When we are sure that our analysis will be limited to the filtered
    # entities, the filtering can be done in place so to limit memory footprint.
    # For example, let's assume that our analysis needs only the first 3
    # successful units. We filter the entities of type 'unit' with state 'DONE'
    # and then select the first three of them. We also sort the units based on
    # their uid before selecting the first three of them:
    ppheader("Filter the first 3 successful 'unit'")
    session.filter(etype=['unit'], state=[rp.DONE])
    units = sorted(session.list('uid'))
    session.filter(uid=units[:3])
    pprint.pprint(session.list(['etype', 'state', 'uid']))

    # Clearly, all this can be done in a one liner. We are nice like that:
    ppheader("Filter the first 3 successful 'unit' - one liner")
    session.filter(etype=['unit'],
                   state=[rp.DONE]).filter(uid=sorted(session.list('uid'))[:3])
    pprint.pprint(session.list(['etype', 'state', 'uid']))

    sys.exit(0)

